
//**************************************************************************
//    PvmStream.h
//
//    The Cor_pvmbuf Class - A buffer class for providing stream access
//    to PVM communication. 
//    (PVM - Parallel Virtual Machine)
//
//**************************************************************************

#ifdef DO_PVM

#ifndef _PVMSTREAM_H_
#define _PVMSTREAM_H_


#include <string.h>
#include <errno.h>
#include <fstream.h>
#include "CoralStream.h"
#include "pvm3.h"


// the file to store the services' pvm tids
// each line of the file contains: <service name> tid
#define PVM_COMM_FILE "/afs/cs.wisc.edu/u/w/h/whuang/reef/.pvm_comm"

// msg tag used by pvm (magic integer)
#define CORAL_TAG 23457


class Cor_pvmbuf : public streambuf {

private:

    int my_tid;                 // the process's own tid
    int rec_tid;                // the receiver's tid
    
    int get_buffer_size;        // the buffer size of the get area
    int put_buffer_size;        // the buffer size of the put area

    Cor_pvmbuf(const Cor_pvmbuf& other);
    Cor_pvmbuf& operator=(const Cor_pvmbuf& other);

protected:

    // consume characters in the put area
    int overflow(int c = EOF);

    // supply characters for fetching. put them into the get area 
    int underflow();          

    // allocate the put and get areas
    int doallocate();

    // initialize the buffer
    void  init();

public:

    // the constructor for a client --> server Cor_pvmbuf 
    Cor_pvmbuf();

    // the constructor for a Cor_pvmbuf with receiver's tid supplied
    Cor_pvmbuf(int _rec_tid);

    // the constructor for a Cor_pvmbuf with both ends' tids supplied
    Cor_pvmbuf(int _my_tid, int _rec_tid);

    // the destructor
   ~Cor_pvmbuf();

    // check if the Cor_pvmbuf is valid
    int Valid();
    
    // close the Cor_pvmbuf. the process is leaving pvm. 
    // only client processes should call it. 
    int close();                

    // synchronize the put area. 
    int sync();	
    
};

//**************************************************************************


#endif   _PVMSTREAM_H_

#endif // DO_PVM
