// *************************************************************************
//    QFModuleData.h
// *************************************************************************

#ifndef _QFModuleData_H_
#define _QFModuleData_H_

#include "RuleInfoArray.h"
#include "ModulePred.h"
#include "ModuleInfo.h"
#include "PredAnnotations.h"
#include "Predicate.h"
#include "RuleInfo.h"
#include "RelationMisc.h"
#include "SemiNaive.h"        // ModuleKind

// *************************************************************************

class Cor_Table;
class Cor_TupleIterator;
class Cor_StorageRelation;
class Cor_VoidStack;

// *************************************************************************
// For each exported query form in a user-defined CORAL module, there is
// one QFModuleData structure created. The Solve() method provides a
// mechanism to evaluate the rules defined in the module. Each such
// structure is associated with its exported query form in a table
// of DerivedMethods (see DerivedRelation.h). QFModuleData is a generic
// class with subclasses for Materialzed and Pipelined modules.
// Sudarshan: This is a static data structure and should not contain
// any fields which differ across different calls to a module.
// *************************************************************************

class Cor_QFModuleData {

  public:
    Cor_Table           *F;             // Table for the query
    Cor_RuleInfoArray   *ruleArray;     // Array of rules for the query form

    Cor_ModulePred      *predicates;    // Used for run_time initialisation
                                        // of rule pointers
    Cor_ModuleInfo      *module_info;
    Cor_PredAnnotations *index_annotations;
    Cor_PredAnnotations *multiset_annotations;
    Cor_PredAnnotations *agg_sel_annotations;
    Cor_PredAnnotations *diskrel_annotations;
    Cor_PredAnnotations *asyncrel_annotations;

    Cor_Predicate       *export_pred;    // Predicate on which query is asked

    RelationOptions      rel_options;
    long                 display_rules;

    Cor_RuleInfo        *curr_rInfo;

    Cor_StorageRelation *magic_rel;       // Saved so we can delete it later
    Cor_Table           *free_table_list; // List of (initialized) table structs
                                          // that can be reused.

    virtual void Solve(Cor_Table *F, Cor_TupleIterator *iterator = NULL) = 0;
    virtual ModuleKind kindof() = 0;

             Cor_QFModuleData();
    virtual ~Cor_QFModuleData();
};

// *************************************************************************

#endif _QFModuleData_H_
