
//**************************************************************************
//    WodiConnectArray.C
//
//    The WodiConnectArray Class - Used to implement connections
//    by Wodi.
//
//    A ConnectArray holds an array of Connections, of which each
//    holds an array of Users.
//     
//**************************************************************************

#ifdef    DO_WODI

#include "Identify.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Assert.h"
#include "WodiConnectArray.h"

//**************************************************************************
//    WodiConnectArray Constructor 
//
//**************************************************************************

Cor_WodiConnectArray::Cor_WodiConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    num_conn = 0;
    cur_conn = 0;
    max_conn = MAX_CONNECTIONS;
    
    for (int i = 0; i < max_conn; i++)
        connArray[i] = (Cor_Connection *) NULL;
    
}

//**************************************************************************
//    WodiConnectArray Destructor -
//    Closes each connection in the array, then shutdown the server.
//    
//**************************************************************************

Cor_WodiConnectArray::~Cor_WodiConnectArray()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Close all connections
    for (int i = 0; i < MAX_CONNECTIONS; i++)
        if (connArray[i])
            delete connArray[i];
    
}

//**************************************************************************
// NextConn() -
//
//**************************************************************************

int Cor_WodiConnectArray::NextConn()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_wodibuf * pbuf = new Cor_wodibuf(WODI_SLAVE);
    CoralStream * cs = new CoralStream(pbuf);
    connArray[num_conn] = new Cor_Connection(cs);
    connArray[num_conn]->conn_id = num_conn;
    cur_conn = num_conn;
    num_conn ++;
    connArray[cur_conn]->useConnection();
    return 1;

}

//**************************************************************************
// DeleteConnection() -
//    This function deletes a connection from the PvmConnectArray object.
//
//**************************************************************************

void Cor_WodiConnectArray::DeleteConnection(int conn_indx)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!connArray[conn_indx] || conn_indx >= num_conn)
        return;

    delete connArray[conn_indx];

    for (int i = conn_indx; i < num_conn-1; i++) {
        connArray[i] = connArray[i+1];
        connArray[i]->conn_id = i;
    }
    connArray[num_conn - 1] = (Cor_Connection *) NULL;        
    
    num_conn --;

}

//**************************************************************************
// ExplainConnection() -
//    This function is callled when a connection request is recieved by the
//    explain server.  This function decides if the connection attempt is
//    valid and accepts or refuses it.
//
//    Return: The return value is zero if everything is okay
//        or < 0 if an error occurs or the connection is refused.
//
//**************************************************************************

int Cor_WodiConnectArray::ExplainConnection()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DO_EXPLAIN
    return 0;

#endif    // DO_EXPLAIN
}

//**************************************************************************

#endif
