// *************************************************************************
//    GenericIndex.C
// *************************************************************************

#include "Globals.h"
#include "GenericIndex.h"

#include "Interface.h"
#include "Debug.h"

// *************************************************************************
int Cor_GenericIndex::matches(Cor_GenericIndex *ind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (argset) {
        if (!ind->argset) return 0;
        if (*argset == *(ind->argset))
            return 1;
        return 0;
    }

    return 1;
}

// *************************************************************************
void Cor_GenericIndex::delete_all_tuples(int)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *(Cor_exEnv->err_strm) << "delete_all_tuples: not implemented yet\n";
}

// *************************************************************************
//  This takes a pattern form index and returns an argument form
//  index, if possible; else it returns NULL.
//
//      (1) if there is a function symbol in pat with an embedded variable 
//          that does not occur in v_nm, return NULL
//      (2) if there is a repeated variable in pat which does not occur 
//          in v_nm, return NULL
// *************************************************************************
char *Cor_simplify_index(Cor_ArgList *pat, int n_var, Cor_ArgList *v_nm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stackmark;
    Cor_BindEnv pat_env(n_var);
    Cor_Term dummy_term(Cor_ZeroCArg, NULL);

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_indexing) {
            *(Cor_exEnv->trace_strm) << "simplify_index: (";
            pat->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "), bound vars = (";
            v_nm->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ")\n";
        }
#    endif

    // First bind each position of pat_env occurring in v_nm to dummy_term
    for (int i=0; i < v_nm->count(); i++) {
        pat_env.bind((*v_nm)[i].var_num(), dummy_term);
    }

    // Now check if each argument of pat is either a variable or a constant
    // in pat_env

    char *return_val = new char[pat->count()+1];
    int j;
    for (j=0; j < pat->count(); j++) {
        if ((*pat)[j].hash(&pat_env) == VarHashValue) {
            if (!(*pat)[j].is_var()){
                // embedded unbound variable
                delete [] return_val;
                stackmark.pop_to();
                return NULL;
            }
            else {
                // free variable
                return_val[j] = 'f';
            }
        }
        else {
            // the argument is ground, given the bound variables
            return_val[j] = 'b';
        }
    }
    return_val[j] = '\0';

    // Now check for repeated unbound variables at top-level in pat.

    for (int j1=0; j1 < pat->count(); j1++){
        if (((*pat)[j1].is_var()) && (return_val[j1] == 'f')) {
            // check if this is a repeated variable
            for (int j2=j1+1; j2 < pat->count(); j2++) {
                if (((*pat)[j2].is_var())
                        && ((*pat)[j1].var_num() == (*pat)[j2].var_num())) {
                      // found a repeated variable
                    delete [] return_val;
                    stackmark.pop_to();
                    return NULL;
                }
            }
        }
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_indexing) {
            *(Cor_exEnv->trace_strm) << 
                "simplify_index: simplified to " <<  return_val << "\n";
        }
#    endif

    stackmark.pop_to();
    return return_val;
}

// *************************************************************************

