// *************************************************************************
//    StorageRelation.C
// *************************************************************************

#include <stdio.h>

#include "StorageRelation.h"

#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "RelationList.h"
#include "RelationMisc.h"
#include "BuiltinSymbols.h"
#include "HashSimpleRelation.h"
#include "Unify.h"
#include "AggSelInfo.h"

#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include "Util.h"
#endif

// *************************************************************************

// temporary hack to avoid changing GenericIndex.h
// also see PersistentIndex.C ::PRAVEEN
#define COR_I_PARTIAL_SUCCESS  2

// *************************************************************************
Cor_StorageRelation *Cor_AllocateRelation(Cor_CArg& name, int arity,
        int delta_indexed)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StorageRelation *rel;
    rel = new Cor_HashSimpleRelation(arity, delta_indexed);
    rel->name = name;
    rel->r_kind = Cor_find_r_kind(name);

#ifdef ASYN_SOCK_REL
    Cor_Relation *r = Cor_find_external_relation(name,arity);
    if (r && !r->_delta_done)
        rel->_delta_done = 0;
    else {
        char tmp[100];
        strcpy(tmp, name.symbol());
        char *ptr = strrchr(tmp, '_');
        if (ptr) {
            int is_adorn = 1;
            for (char *ptr3=ptr+1; *ptr3; ptr3++)
                if ((*ptr3 != 'b') && (*ptr3 != 'f')) {
                    is_adorn = 0;
                    break;
                }

            if (is_adorn)
               *ptr = '\0'; // get rid of the trailing part after '_'

            r = Cor_find_external_relation(Cor_CArg(tmp), arity);
            if (r && !r->_delta_done)
                rel->_delta_done = 0;
        }
    }
#endif
    
    return rel;
}

// *************************************************************************
Cor_StorageRelation::Cor_StorageRelation()
            : Cor_Relation(), headMark()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char namep[100];
    sprintf(namep,"Rel_%d", Cor_RelCount++);
    name = Cor_CArg(namep);
    _arity = 0;

    for_supp_index = NULL;

    r_kind = COR_R_ANSWER;  // default

    delta_relations = NULL;
    free_reln_list  = NULL;

    // initialize circular list of marks
    headMark.next = &headMark;
    headMark.prev = &headMark;

    // to allow reuse of marks
    free_marks = NULL;
}

// *************************************************************************
Cor_StorageRelation::Cor_StorageRelation(int new_arity, int delta)
    : Cor_Relation(new_arity, delta), headMark()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char namep[100];

    sprintf(namep,"Rel_%d", Cor_RelCount++);
    name = Cor_CArg(namep);
    _arity = new_arity;

    for_supp_index = NULL;

    delta_relations = NULL;
    free_reln_list  = NULL;

    // initialize circular list of marks
    headMark.next = &headMark;
    headMark.prev = &headMark;

    free_marks = NULL;    /* to allow reuse of marks */
}

// *************************************************************************
int Cor_StorageRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    count++;
    return(insert_raw(tuple));
}


// *************************************************************************
int Cor_StorageRelation::insert_raw(Cor_Tuple *)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CORAL_error(COR_BAD_TUPLE_INSERT,name.symbol(),"insert_raw");
    ASSERT(0);
    return(-1);
}

// *************************************************************************
//  For now, lets assume that any delta relation that are indexed are
//  HashSimpleRelation's
// *************************************************************************
Cor_RMark *Cor_StorageRelation::getMark()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RMark *mark;

    if (!delta_indexed) {
        if (free_marks) {
            mark = free_marks;
            free_marks = free_marks->next;
            mark->init(&tlist,tlist.count());
        } else {
            mark = new Cor_RMark(&tlist,tlist.count());
        }
    } else {
        Cor_HashSimpleRelation *newreln;
        int create_relation = 1;

        if (!delta_relations) {
            delta_relations = new Cor_RelationList();
        } else {
            /*
            NOTE: Restore this after adding counts to all relations.
             if (delta_relations->reln_lastptr &&
                     (delta_relations->reln_lastptr->reln->count == 0)) {
              // there's no need to create a new reln.
              // Just do mark->init(delta_relations->reln_lastptr);
                 create_relation = 0;
             }
            */
        }

        if (create_relation) {
            if (!free_reln_list) {

                newreln = new Cor_HashSimpleRelation(arity(), delta_indexed,
                    Cor_exEnv->C_delta_buckets_default);

                // NOTE: This is to circumvent a problem with the hashing
                // code, which uses the relation name unnecessarily.
                newreln->name = name;
                add_all_indices(newreln);
                delta_relations->append_relation(newreln);
            } else {
                // Reuse relation in list
                Cor_RelationLink *new_reln_link = free_reln_list;
                free_reln_list = free_reln_list->next;
                new_reln_link->next = NULL;
                delta_relations->append(new_reln_link);
            }
        }

        if (free_marks) {
            mark = free_marks;
            free_marks = free_marks->next;
            mark->init(delta_relations->reln_lastptr);
        } else {
            mark = new Cor_RMark(delta_relations->reln_lastptr);
        }
    }

    // Link it in
    mark->next = &headMark;
    mark->prev = headMark.prev;
    headMark.prev->next = mark;
    headMark.prev = mark;
    return mark;
}

// *************************************************************************
void Cor_StorageRelation::freeMark(Cor_RMark *mark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete mark;
}

// *************************************************************************
void Cor_StorageRelation::free_allMarks()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RMark *mark, *prev = NULL;

    for (mark = free_marks; mark; mark = mark->next) {
        delete prev;
        prev = mark;
    }

    delete prev;
}

// *************************************************************************
// to be used internally, not by user
int Cor_StorageRelation::add_index(Cor_BitVector *bv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_GenericIndexSet *index_set;
    Cor_GenericIndex *h;
    int ret_val;

    if ((index_set = get_index_set()) == NULL) {
      *(Cor_exEnv->err_strm) << 
          "StorageRelation::add_index found NULL index_set\n";
      return(COR_I_FAIL);
    }

    if (( h = create_index_object(bv)) == NULL) {
        *(Cor_exEnv->err_strm) << 
            "StorageRelation::add_index unable to create index object\n";
        return(COR_I_FAIL);
    }

    if ((ret_val = index_set->insert_index(h)) == COR_I_FAIL) {
        // temporary hack : need to add reclaim_index_object method !!
        //*(Cor_exEnv->err_strm)
        //      << "StorageRelation::add_index unable to insert index\n";
        Cor_reclaim_index(h, r_kind);
        return(COR_I_FAIL);
    }
    else if (ret_val == COR_I_PARTIAL_SUCCESS) {
        return(COR_I_PARTIAL_SUCCESS);
    }

    index_set->fill_index(h);

    // add the same index to the relation pointed to by for_supp_index
    if (for_supp_index != NULL) {
        for_supp_index->add_index(bv);
    }

    // add the same index to each of the delta_indexed relations
    if (delta_indexed && delta_relations) {
        for (Cor_RelationLink *tmp=delta_relations->reln_chain; tmp != NULL;
                tmp=tmp->next) {
            tmp->reln->add_index(bv);
        }
    }

    return(COR_I_SUCCESS);
}

// *************************************************************************
int Cor_StorageRelation::add_index(char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(adorn != NULL);

    Cor_BitVector *bv = new Cor_BitVector(arity());

    char *adorn1 = adorn;
    int i;
    for (i = 0; i < arity(); i++, adorn++) {
        if (*adorn == 0) {
            *(Cor_exEnv->err_strm) << 
                "StorageRelation::add_index adorn too short: \n";
            *(Cor_exEnv->err_strm) << "    --- arity of "
                << name.symbol() << "=" << arity()
                << ", adorn len=" << i << ")\n";
            delete bv;
            return(COR_I_FAIL);
        }
        if ((*adorn == 'b') || (*adorn == '1'))
            bv->set(i);
    }

    if (*adorn != '\0') {
        *(Cor_exEnv->err_strm) << 
                "StorageRelation::add_index adorn too long\n";
        *(Cor_exEnv->err_strm) << "    --- arity of "
            << name.symbol() << "=" << arity()
            << ", adorn len=" << i << " adorn = " << adorn1 << ")\n";
        delete bv;
        return(COR_I_FAIL);
    }

    int ret = this->add_index(bv);
    delete bv;
    return(ret);
}

// *************************************************************************
int Cor_StorageRelation::add_index(Cor_ArgList *args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT(args->count() != 0);

    if (args->count() > 1) {
        *(Cor_exEnv->err_strm) << "More than one argument to add_index "
                << "-- using first one to build index.\n";
    }

    ASSERT((*args)[0].is_symbol());
    char *arg_str = ((*args)[0]).symbol();
    return add_index(arg_str);
}

// *************************************************************************
int Cor_StorageRelation::add_index(Cor_ArgList *pat, int n_var,
        Cor_ArgList *v_nm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *argindex = Cor_simplify_index(pat, n_var, v_nm);

    if (argindex)
        return add_index(argindex);

    Cor_GenericIndexSet *index_set;
    Cor_GenericIndex *h;

    if ((h  = create_index_object(pat, n_var, v_nm)) == NULL)
        return(COR_I_FAIL);

    if ((index_set = get_index_set()) == NULL)
        return(COR_I_FAIL);

    if (index_set->insert_index(h) == COR_I_FAIL)
        return(COR_I_FAIL);

    index_set->fill_index(h);

    // add the same index to the relation pointed to by for_supp_index
    if (for_supp_index != NULL) {
        for_supp_index->add_index(pat, n_var,v_nm);
    }

    // add the same index to each of the delta_indexed relations
    if (delta_indexed && delta_relations) {
        for (Cor_RelationLink *tmp=delta_relations->reln_chain; tmp != NULL;
            tmp=tmp->next)
        {
            tmp->reln->add_index(pat, n_var, v_nm);
        }
    }
    return(COR_I_SUCCESS);
}


// *************************************************************************
/*
    Called to add all the indices in one relation on to another (usually
    on to a newly created delta relation.
*/

int Cor_StorageRelation::add_all_indices(Cor_StorageRelation *reln)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_GenericIndexSet *index_set;
    if ((index_set = this->get_index_set()))
        return(index_set->add_all_indices(reln));
    else
        return(0);
}

// *************************************************************************
/*
    Called to add all the tuples in (this) into reln.
*/

int Cor_StorageRelation::add_all_tuples(Cor_StorageRelation *reln,
    int marks_into_account, int check_dupl)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_GenericIndexSet *index_set;

    if ((index_set = this->get_index_set())) {
        return index_set->add_all_tuples(reln, marks_into_account,
            check_dupl);
    } else
        return(0);
}

// *************************************************************************
/*
    Called to convert all tuples in (this) into goal nodes and
    add them to reln and context.  For OrdSearch
*/

int Cor_StorageRelation::add_all_goals(Cor_StorageRelation *reln,
    Cor_Context *context, Cor_StorageRelation *owner_rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_GenericIndexSet *index_set;

    if ((index_set = this->get_index_set())) {
        return index_set->add_all_goals(reln, context, owner_rel);
    } else
        return(0);
}

// *************************************************************************
void Cor_StorageRelation::add_agg_sel_info(Cor_AggSelInfo *newinfo)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // first check that this aggsel has not already been specified
    if (Cor_duplicate_agg_sel_anno(this, newinfo)) {
        delete newinfo;
    } else {
        newinfo->next = agg_sel_info;
        agg_sel_info = newinfo;
    }
}

// *************************************************************************
void Cor_StorageRelation::add_agg_sel_anno(Cor_PredAnnotations *anno)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_CArg *op_arg = &(anno->op_arg);
    int arity = 0;
    CArgKind agg_kind = EMPTY_ARG;
    Cor_ArgList *args = NULL;

    switch (op_arg->Kind()) {

    default:
        Cor_annotation_error(anno, "Bad aggregation type");
        return;

    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:
    case STDDEV_GRPING_ARG:
    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:
    case SET_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
          // possibly extend to irrelevant fact annotation
        Cor_annotation_error(anno,
            "Only min(), max(), and any() aggregation handled.");
        break;

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case ANY_GRPING_ARG:
        agg_kind = op_arg->Kind();
        arity = op_arg->grp_arity();
        args  = &(op_arg->nested_args()->args());
        break;

    case FUNCT_ARG:
        arity = op_arg->arity();
        args  = &(op_arg->args());

        if (op_arg->nm_equals(*Cor_MaxOpSymbol))
            agg_kind = MAX_GRPING_ARG;
        else if (op_arg->nm_equals(*Cor_MinOpSymbol))
            agg_kind = MIN_GRPING_ARG;
        else if (op_arg->nm_equals(*Cor_AnyOpSymbol))
            agg_kind = ANY_GRPING_ARG;
        else {
            agg_kind = USR_DEFD_GRPING_ARG;
        }

        break;
    }

    if (agg_kind != USR_DEFD_GRPING_ARG) {

        if ((arity != 1)
                && (agg_kind != ANY_GRPING_ARG)) {
            Cor_annotation_error(anno,
                "Bad number of arguments for aggregate selection.");
            return;
        }

        for (int j=0; j < arity; j++) {
    
            Cor_AggSelInfo *newinfo = new Cor_AggSelInfo();
    
            newinfo->agg_kind = agg_kind;
            newinfo->op_arg = op_arg->dup();
            newinfo->agg_type = anno->agg_type;
    
            Cor_ArgList *arglist = newinfo->arglist = anno->arglist1;
    
            newinfo->agg_val = new Cor_ArgList( (*args)[j] );
    
            if (! (*(newinfo->agg_val))[0].is_var()) {
                Cor_annotation_error(anno,
                        "Aggregated argument should be a variable.");
                delete newinfo;
                return;
            }
    
            newinfo->var_count = arglist->max_vars();
    
            Cor_ArgList *copy_args = new Cor_ArgList(arglist->arity());
            ASSERT(copy_args);
            Cor_IntStack *groupby_vars = anno->arglist2->var_list(NULL);
    
            Cor_BitVector *groupargs = new Cor_BitVector(arglist->arity());
            groupargs->clear();
    
            for (int i=0; i < arglist->arity(); i++) {
                if ( !(*arglist)[i].is_var() ) {
                    Cor_annotation_error(anno,
                            "Arglist should contain only variables");
                    delete newinfo;
                    delete copy_args;
                    delete groupargs;
                    return;
                }
    
                Cor_CArg *oldvar = &((*arglist)[i]);
    
                // Search for var in anno->arglist2
                if ( groupby_vars->contains(oldvar->var_num()) ) {
                    (*copy_args)[i] = *oldvar;
                    groupargs->set(i);
                } else {
                    (*copy_args)[i].set_var(newinfo->var_count++);
                }
    
                if (oldvar->var_num() == (*(newinfo->agg_val))[0].var_num()) {
                    newinfo->gb_agg_val = new Cor_ArgList( (*copy_args)[i] );
                }
            }
    
            delete groupby_vars;
    
            newinfo->groupbylist = copy_args;
            newinfo->bindenv = new Cor_BindEnv(newinfo->var_count);
    
            // Add the aggregate selection, and create an index too
            // WARNING: Better to move the index generation into the rewriting
            // phase, so that it is visible.
    
            add_agg_sel_info(newinfo);
            add_index(groupargs);
    
            delete groupargs;
        }

    } else {
        // User-defined grouping

        Cor_AggSelInfo *newinfo = new Cor_AggSelInfo();

        newinfo->agg_kind = agg_kind;
        newinfo->op_arg = op_arg->dup();
        newinfo->agg_type = anno->agg_type;

        Cor_ArgList *arglist = newinfo->arglist = anno->arglist1;

        newinfo->agg_val = op_arg->args().dup();
        newinfo->gb_agg_val = newinfo->agg_val->dup();

        newinfo->var_count = arglist->max_vars();

        Cor_ArgList *copy_args = new Cor_ArgList(arglist->arity());
        ASSERT(copy_args);
        Cor_IntStack *groupby_vars = anno->arglist2->var_list(NULL);

        Cor_BitVector *groupargs = new Cor_BitVector(arglist->arity());
        groupargs->clear();

        for (int i=0; i < arglist->arity(); i++) {
            if ( !(*arglist)[i].is_var() ) {
                Cor_annotation_error(anno,
                        "Arglist should contain only variables");
                delete newinfo;
                delete copy_args;
                delete groupargs;
                return;
            }

            Cor_CArg *oldvar = &((*arglist)[i]);

            // Search for var in anno->arglist2
            if ( groupby_vars->contains(oldvar->var_num()) ) {
                (*copy_args)[i] = *oldvar;
                groupargs->set(i);
            } else {
                (*copy_args)[i].set_var(newinfo->var_count++);
            }

            for (int m=0; m < newinfo->gb_agg_val->arity(); m++)
                if ( (*(newinfo->gb_agg_val))[m].is_var()
                     && (oldvar->var_num()
                            == (*(newinfo->gb_agg_val))[m].var_num())) {
                (*(newinfo->gb_agg_val))[m] =  (*copy_args)[i];
            }
        }

        delete groupby_vars;

        newinfo->groupbylist = copy_args;
        newinfo->bindenv = new Cor_BindEnv(newinfo->var_count);

        // Add the aggregate selection, and create an index too
        // WARNING: Better to move the index generation into the rewriting
        // phase, so that it is visible.

        add_agg_sel_info(newinfo);
        add_index(groupargs);

        delete groupargs;
    }
}

// *************************************************************************
//  insert:  returns 1 if a tuple was inserted into the end of a relation.
//  In the case of aggregation this function calls insert_grouping, and the
//  convention  is more complicated -
//      a) returns 1 if the new tuple is inserted
//      b) returns 0 in the case that an existing tuple was modified, but
//          new tuple not inserted.
//      
//  SPF: - modification
//    Have changed b) to return 1 if the tuple is modified, 0 if not.
//    This causes a problem with leaking inserted tuples, but gives
//    the correct results in the case of "declad9".
// *************************************************************************
int Cor_StorageRelation::insert(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            *(Cor_exEnv->trace_strm)
                    << "[Inserting " << name.symbol() << "(";
            tuple->args().print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ")";
        }
#    endif

    if (tuple->arity() != arity()) {
        *(Cor_exEnv->err_strm)
            << "Relation::insert - attempt to insert tuple of arity "
            << tuple->arity() << endl
            << "                 - in relation "
            << name.symbol()
            << " of arity " << arity() << endl;
        return(0);
    }

    if (tuple->args().has_grouping_args())
        return insert_grouping(tuple->args(),NULL);

    return(insert_tuple(tuple));
}

// *************************************************************************
int Cor_StorageRelation::insert_grouping(Cor_ArgList& args, Cor_BindEnv *env)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark0;

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS) {
            *(Cor_exEnv->trace_strm)
                << "[Insert grouping: %s(", name.symbol();
            args.print(env, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ")";
        }
#    endif

    int i;

    Cor_ArgList *temp_args = new Cor_ArgList(args.count());

    // WARNING::  ReturnUnify optimizations needed here!!!
    // See how this is affected by and will affect return unify optimizations

    for (i=0; i < args.count(); i++) {

        if (args[i].is_grouping())

            // Replace DontCareArg by a new variable !!?????
            // Tricky replacement - unification on args will affect
            // grouping arg in case it shares variables with
            // the non-grouping arg.

            (*temp_args)[i] = *Cor_TheDontCareArg;

        else
            (*temp_args)[i] = args[i];
    }

    Cor_TupleIterator iter(this, *temp_args, env);
    Cor_StackMark stack_mark1;
    int do_insert = 1;
    int stat=0;

    for (;;) {
        Cor_StackMark stackmark2;
        Cor_Tuple *tuple1 = iter.get_next_tuple();

        // Undo any bindings done by get_next_tuple();
        stackmark2.pop_to();

        if (tuple1 == NULL)
            break;

        if (Cor_subsumes_arg_lists(*temp_args, tuple1->args(),
                    env, iter.tuple_env)) {
#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS) {
                    *(Cor_exEnv->trace_strm)
                        << " - non-grouping args subsume - (";
                    tuple1->printon(Cor_exEnv->trace_strm);
                    *(Cor_exEnv->trace_strm) << ")]\n";
                }
#            endif
#            ifdef DO_EXPLAIN
                char copy_arglist[5000];
                if (this->local_options & REL_EXPLAIN) {
                    int pos = 0;
                    (tuple1->args()).sprint(copy_arglist, &pos);
                }
#            endif

#            ifdef DO_EXPLAIN
            Cor_ArgList *expln_args = tuple1->_args;
#            endif

            do_insert = 0;
            stat = Cor_update_grouping(this, tuple1, &args, env);

#            ifdef DO_EXPLAIN
                if (this->local_options & REL_EXPLAIN) {
                    Cor_group_dump(env, expln_args, this, copy_arglist);
                }
#            endif

        }
        stack_mark1.pop_to();
        break;
    }
    stack_mark0.pop_to();

    delete temp_args;

    if (! do_insert) {
#        ifdef DO_TRACE
            if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
                if (!DISPLAY_INSERTIONS) {
                    *(Cor_exEnv->trace_strm)
                        << "[Inserting Grouping " << name.symbol() << "(";
                    args.print(env, Cor_exEnv->trace_strm);
                }
                *(Cor_exEnv->trace_strm) <<
                  ") - no new fact created (old one may have been modified)\n";
            }
#        endif

        return(stat);
    }

    Cor_ArgList *copy_args = new Cor_ArgList(arity());

    // CopyArgs also renames variables so that they are distinct,
    // and the bindenv need not be used to distinguish variables.
    // Grouping::simplify does the ``right'' thing with aggregation -
    // it removes the grouping junk and returns the argument.
    // In fact, for count it just returns a default arg since it
    // will never be looked at.
#ifdef    DEBUG
    int env_size =
#endif
        Cor_CopyArgs(copy_args, &args, env, arity(), Cor_UnusedEnv);

      // This is the case if the result is fully ground
    ASSERT(env_size == 0);

/*
    Cor_ArgList *copy_args = args.dup();

    int is_const = copy_args->simplify_me(env);

    if (!is_const) {
        *(Cor_exEnv->err_strm) << "Error: Non-ground fact insert: "
                               << *copy_args << endl;
       
        delete copy_args; 
        return(0);
    }
*/

#    ifdef DO_TRACE
        if (DISPLAY_INSERTIONS | DISPLAY_NEW_INSERTIONS) {
            *(Cor_exEnv->trace_strm) << 
                "[Inserting Grouping " <<  name.symbol() << "(";
            copy_args->print(NULL, Cor_exEnv->trace_strm);
            if (env_size == 0)
                *(Cor_exEnv->trace_strm) << " - done (ground)]\n";
            else
                *(Cor_exEnv->trace_strm) << 
                            " - done (env-size:" <<  env_size << ")]\n";
        }
#    endif

    // the new tuple created has copied the arglist.  subsequent
    // modifications to the tuple by Grouping::update_tuple
    // can hence update the arglist in place without worrying
    // about it being shared by anyone else

    Cor_Tuple *tuple2 = new Cor_Tuple(copy_args);

    stat = insert_tuple(tuple2);

#    ifdef DO_EXPLAIN
        if (this->local_options & REL_EXPLAIN)
            Cor_group_dump(env, tuple2->_args, this);
#    endif

      // This really can only happen for a relation that does not
      // actually store the tuple in memory
      // (socketrel or a persistent relation perhaps)
    if (!stat) {
        delete copy_args;
        delete tuple2;
    }

    return(1);
}

// *************************************************************************
int Cor_StorageRelation::is_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark,
        Cor_RMark *endmark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StackMark stack_mark1;

    // Check for ordinary subsumption
    Cor_TupleIterator iter(this, *(tuple->_args), NULL, NULL,
        startmark, endmark);
    for (;;) {
        Cor_StackMark stackmark2;
        Cor_Tuple *tuple1 = iter.get_next_tuple();
          // Undo any bindings done by get_next_tuple();
        stackmark2.pop_to();

        if (tuple1 == NULL)
            break;
        if (Cor_subsumes_arg_lists(tuple1->args(), *(tuple->_args),
                iter.tuple_env, NULL)) {
#            ifdef DO_TRACE
                if (DISPLAY_INSERTIONS) {
                    *(Cor_exEnv->trace_strm) << " - subsumed by (";
                    tuple1->printon(Cor_exEnv->trace_strm);
                    *(Cor_exEnv->trace_strm) << ")]\n";
                }
#            endif
            stack_mark1.pop_to();
            return(1);
        }
        stack_mark1.pop_to();
    }
    stack_mark1.pop_to();

    return is_agg_subsumed(tuple, startmark, endmark);
}

// *************************************************************************

#define GETBINDENV(aggsel,env) \
  { \
    env = aggsel->bindenv;     \
    aggsel->bindenv = NULL;  /*To allow reentrant use of aggsel->bindenv*/ \
    if(env==NULL) env = new Cor_BindEnv(aggsel->var_count);           \
  }

#define RELEASE(aggsel,env) \
  { \
    if (aggsel->bindenv==NULL) aggsel->bindenv = env; \
    else delete env;        \
  }

// *************************************************************************
int Cor_StorageRelation::is_agg_subsumed(Cor_Tuple *tuple, Cor_RMark *startmark,
    Cor_RMark *endmark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k;
    // Check aggregate selections

    // Quick check that this call is even needed.
    if (!agg_sel_info)
        return(0);

    // WARNING::  ReturnUnify optimizations needed here!!!

    Cor_StackMark stack_mark1;
    Cor_AggSelInfo *aggsel;

    for (aggsel=agg_sel_info; aggsel != NULL; aggsel=aggsel->next) {

          // These are done only once per iteration, and are
          // handled elsewhere.
        if ((aggsel->agg_kind == USR_DEFD_GRPING_ARG)
                && (aggsel->agg_type == AggSel_PerIter))
            continue;

        // For each agg sel in the list
        Cor_StackMark stack_mark2;
        Cor_BindEnv *aggsel_env;
        GETBINDENV(aggsel, aggsel_env);

        Cor_unify_arg_lists(*(aggsel->arglist), *(tuple->_args), aggsel_env,
            NULL);
        Cor_TupleIterator iter(this, *(aggsel->groupbylist), aggsel_env,
            NULL, startmark, endmark, &(aggsel->indexnum));

        switch (aggsel->agg_kind) {

        case MIN_GRPING_ARG:
        case MAX_GRPING_ARG:
        case ANY_GRPING_ARG:

            for (;;) {
                Cor_StackMark stackmark3;
                Cor_Tuple *dbtuple = iter.get_next_tuple();
    
                // aggsel->groupbylist and aggsel->gb_agg_val correspond to
                // the fetched tuple.
                stackmark3.pop_to();
    
                if (iter.no_match() || (dbtuple == NULL))
                    break;
    
                if (Cor_subsumes_arg_lists(*(aggsel->groupbylist),
                    dbtuple->args(), aggsel_env, iter.tuple_env)) {
    
                  // Check if new value (aggsel->agg_val) subsumes old
                  // value (aggsel->gb_agg_val).
                    int stat = Cor_subsumes_value(aggsel->agg_kind,
                          (*(aggsel->agg_val))[0],
                          (*(aggsel->gb_agg_val))[0],
                          aggsel_env, aggsel_env);
                    if (stat > 0) {
#                 ifdef DO_TRACE
                          if (Cor_exEnv->dbg_aggregates || DISPLAY_INSERTIONS) {
                                *(Cor_exEnv->trace_strm) << 
                                  "\n\t\t - aggsel: new fact subsumes";
                                *(Cor_exEnv->trace_strm) << ": "
                                      << name.symbol() << "(";
                                dbtuple->printon(Cor_exEnv->trace_strm);
                                *(Cor_exEnv->trace_strm) << ")\n";
                          }
#                 endif
                      delete_id_tuple(this, dbtuple);
                    } else if (stat < 0) {
#                 ifdef DO_TRACE
                          if (Cor_exEnv->dbg_aggregates || DISPLAY_INSERTIONS) {
                                *(Cor_exEnv->trace_strm) << 
                                    "\n\t\t - aggsel: old fact subsumed by";
                                *(Cor_exEnv->trace_strm) << ": "
                                      << name.symbol() << "(";
                                dbtuple->printon(Cor_exEnv->trace_strm);
                                *(Cor_exEnv->trace_strm) << ")\n";
                          }
#                 endif
                      stack_mark1.pop_to();
                      RELEASE(aggsel,aggsel_env)
                      return(1);
                  }
                    // New fact not subsumed
                    // using current aggregate selection
                }
                stackmark3.pop_to();
             }
             break;

        case USR_DEFD_GRPING_ARG: {

                  // Locate the requested aggregate selector
                if (!aggsel->rel) {

                    aggsel->rel = find_relation(
                        aggsel->op_arg->functor(), AGG_SELECTOR_ARITY);

                    if (!aggsel->rel) {
                        *(Cor_exEnv->err_strm)
                             << "Error: Aggregate Selector "
                             << aggsel->op_arg->functor()
                             << ", not found.\n";
                        break;
                    }

                    // check arity as find_relation does not
                    if (aggsel->rel->arity() != AGG_SELECTOR_ARITY) {
                        *(Cor_exEnv->err_strm)
                            << "Error: Invalid Aggregate Selector "
                            << aggsel->op_arg->functor() << ", wrong arity.\n";
                        break;
                    }
                }

                if (aggsel->agg_type == AggSel_Regular) {

                    Cor_TupleStack grping_stk, answer_stk;
                    Cor_Tuple *dbtuple;
    
                    for (;;) {
                        Cor_StackMark stackmark3;
                        dbtuple = iter.get_next_tuple();
    
                        stackmark3.pop_to();
    
                        if (iter.no_match() || (dbtuple == NULL))
                            break;
    
                        grping_stk.push(UnknownHashValue,dbtuple);
                    }
    
                    Cor_ArgList *al = new Cor_ArgList(AGG_SELECTOR_ARITY);
                    Cor_CArg *var = new Cor_CArg();
                    Cor_BindEnv *env = new Cor_BindEnv(1);
                    var->set_var(0);
    
                    Cor_CArg *a0, *a1, *a2, *a3;
    
                    if (grping_stk.is_empty()) {
                          // Relation was previously empty
                        dbtuple = NULL;
                        a0 = new Cor_CArg(NIL_ARG);
                        a1 = new Cor_CArg(NIL_ARG);
                    } else {
                        dbtuple = grping_stk.get_ith_tuple(0);
    
                        if (Cor_subsumes_arg_lists(*(aggsel->groupbylist),
                                dbtuple->args(), aggsel_env, NULL)) {
    
                            a0 = dbtuple->args().make_list();
                            a1 = aggsel->gb_agg_val->make_list();
                        } else {
                            a0 = new Cor_CArg(NIL_ARG);
                            a1 = new Cor_CArg(NIL_ARG);
                        }
                    }
    
                    a2 = tuple->args().make_list();
                    a3 = aggsel->agg_val->make_list();
    
                    (*al)[0] = *a0; 
                    (*al)[1] = *a1; 
                    (*al)[2] = *a2; 
                    (*al)[3] = *a3; 
    
                    delete a0;
                    delete a1;
                    delete a2;
                    delete a3;
    
                    al->simplify_me(aggsel_env);
                    (*al)[4] = *var;
    
                    Cor_TupleIterator *it = 
                        new Cor_TupleIterator(aggsel->rel, *al, env);
    
                    it->get_next_tuple();
    
                    var->simplify_me(env);
                    var->hash();
    
                    if (it->no_match() || var->is_nil()) {
    
                        // No answers
    
                        // Delete all answers already in database
                        for (j=0; j < grping_stk.count(); j++) {
                            dbtuple = grping_stk.get_ith_tuple(j);
                            delete_id_tuple(this,dbtuple);
                        }
    
                        delete al;
                        delete var;
                        delete it;
    
                        stack_mark1.pop_to();
                        delete env;
                        RELEASE(aggsel,aggsel_env)

                        // new tuple is subsumed
                        return(1);
                    } else {
    
                        if (!var->is_list()) {
                            *(Cor_exEnv->err_strm)
                                << "Error: Aggregate Selector "
                                << aggsel->rel->name
                                << " did not return a list.\n";
                            break;
                        }
    
                        Cor_Tuple *anstuple = var->unmake_list();
    
                        if (! anstuple->equals(dbtuple)) {
                            if (dbtuple)
                                delete_id_tuple(this, dbtuple);
    
                            if (!insert_tuple(anstuple))
                                delete anstuple;
    
                            anstuple = NULL;
                        }
    
                        delete anstuple;
                    }
    
                    delete it;
                    delete al;
                    delete var;
    
                    stack_mark1.pop_to();
                    delete env;
                    RELEASE(aggsel,aggsel_env)

                      // new tuple is subsumed, but different one was added
                    return(-1);
                } else {
                      // AggSel_MultAnswers
                    ASSERT(aggsel->agg_type == AggSel_MultAnswers);

                    Cor_TupleStack grping_stk, answer_stk;
        
                    for (;;) {
                        Cor_StackMark stackmark3;
                        Cor_Tuple *dbtuple = iter.get_next_tuple();
            
                        // aggsel->groupbylist and aggsel->gb_agg_val
                        // correspond to the fetched tuple.
                        stackmark3.pop_to();
            
                        if (iter.no_match() || (dbtuple == NULL))
                            break;
        
                        grping_stk.push(UnknownHashValue,dbtuple);
                    }
       
                    Cor_ArgList *al
                            = new Cor_ArgList(AGG_SELECTOR_ARITY);
                    Cor_CArg *var = new Cor_CArg();
                    var->set_var(0);

                    Cor_CArg *a0, *a1, *a2, *a3, *tmparg;
                    a0 = new Cor_CArg(NIL_ARG);
                    a1 = new Cor_CArg(NIL_ARG);

                    aggsel_env->clear_all();
                    Cor_unify_arg_lists(*(aggsel->arglist),
                            *(tuple->_args), aggsel_env, NULL);

                    a2 = tuple->args().make_list();
                    a3 = aggsel->agg_val->make_list();
                    a3->simplify_me(aggsel_env);
        
                    for (k=0; k < grping_stk.count(); k++) {
        
                        Cor_Tuple *grp_tuple = grping_stk.get_ith_tuple(k);
       
                        aggsel_env->clear_all();
                        Cor_unify_arg_lists(*(aggsel->arglist),
                                *(tuple->_args), aggsel_env, NULL);
        
                        if (Cor_subsumes_arg_lists(*(aggsel->groupbylist),
                                grp_tuple->args(), aggsel_env, NULL)) {
        
                            tmparg = grp_tuple->args().make_list();
                            tmparg->simplify_me(aggsel_env);
                            a0 = new Cor_CArg(tmparg,a0);
        
                            tmparg = aggsel->gb_agg_val->make_list();
                            tmparg->simplify_me(aggsel_env);
                            a1 = new Cor_CArg(tmparg,a1);
                        }
                    }
        
                    (*al)[0] = *a0; 
                    (*al)[1] = *a1; 
                    (*al)[2] = *a2; 
                    (*al)[3] = *a3; 
        
                    delete a0;
                    delete a1;
                    delete a2;
                    delete a3;
        
                    (*al)[4] = *var;
       
                    Cor_BindEnv *env = new Cor_BindEnv(1);
                    Cor_TupleIterator *it = 
                        new Cor_TupleIterator(aggsel->rel, *al, env);
        
                    it->get_next_tuple();
        
                    while(!it->no_match()) {
        
                        var->simplify_me(env);
                        var->hash();
        
                        if (!var->is_list) {
                            *(Cor_exEnv->err_strm)
                                << "Warning: Aggregate Selector "
                                << aggsel->rel->name
                                << " did not return a list: "
                                << *var << endl;
                            continue;
                        }        
                  
                        var->unmake_lists(&answer_stk);
       
                        var->set_var(0);
                        it->get_next_tuple();
                    }
        
                    delete it;
                    delete al;
                    delete var;
        
                      // MultAnswers

                    for (k=0; k < grping_stk.count(); k++) {
        
                        Cor_Tuple *grp_tuple = grping_stk.get_ith_tuple(k);
        
                        if (! answer_stk.delete_eq_tuple(grp_tuple)) {
                              // This tuple is not in the answer,
                              // remove it from the relation.
                            delete_id_tuple(this,grp_tuple);
                        }
                    }

                    int is_subsumed = 1;
        
                    if (answer_stk.delete_eq_tuple(tuple))
                        is_subsumed = 0;

                        // Answers to be inserted
                    if (!answer_stk.is_empty()) {
                        for (k=0; k < answer_stk.count(); k++) {

                            Cor_Tuple *tuple = answer_stk.get_ith_tuple(k);
        
                              // Insert this answer into the relations
                            if (tuple && !insert_tuple(tuple))
                                delete tuple;
                        }

                          // -1 means insert tuple is subsumed, but
                          // other tuples were added.
                        if (is_subsumed)
                            is_subsumed = -is_subsumed;
                    }

                    stack_mark1.pop_to();
                    delete env;
                    RELEASE(aggsel,aggsel_env)

                    return(is_subsumed);
                }
            }
            break;

        default:
#ifdef    DEBUG
            *(Cor_exEnv->err_strm) << "Unhandled case in "
                                   << __FUNCTION__ << endl;;
#endif
            break;
        }

        stack_mark2.pop_to();
        RELEASE(aggsel,aggsel_env)
    }

    stack_mark1.pop_to();

    // Fact is not subsumed
    return(0);
}

// *************************************************************************
//  For aggregate selection only occuring once per (SemiNaive) iteration.
//  The delta is used to decide which groupings need to be examined/updated. 
//  Return the net number of tuples added/removed from the delta.
// *************************************************************************
int Cor_StorageRelation::per_iteration_aggregation(Cor_RMark *startmark,
    Cor_RMark *endmark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k;
    int num_added = 0;
    Cor_AggSelInfo *aggsel;
    Cor_StackMark   stack_mark1;

    for (aggsel=agg_sel_info; aggsel != NULL; aggsel=aggsel->next) {

          // This is the only type of Aggregate Selection done
          // per iteration, rather than per insert.
        if (aggsel->agg_type == AggSel_PerIter) {

            Cor_StackMark stack_mark2;

#ifdef    AGGSEL_VERBOSE
            if (Cor_exEnv->C_verbose_aggsel_default)
                cout << "Performing an aggregate selection on "
                     << name << endl;
#endif
              // Found a per_iteration aggregate selector.

              // Locate the requested aggregate selector
            if (!aggsel->rel) {

                aggsel->rel = find_relation(
                    aggsel->op_arg->functor(), AGG_SELECTOR_ITER_ARITY);

                if (!aggsel->rel) {
                    *(Cor_exEnv->err_strm)
                         << "Error: Aggregate Selector "
                         << aggsel->op_arg->functor()
                         << ", not found.\n";
                    break;
                }

                // check arity as find_relation does not
                if (aggsel->rel->arity() != AGG_SELECTOR_ITER_ARITY) {
                    *(Cor_exEnv->err_strm)
                        << "Error: Invalid Aggregate Selector "
                        << aggsel->op_arg->functor() << ", wrong arity.\n";
                    break;
                }
            }

            Cor_BindEnv *aggsel_env;
            GETBINDENV(aggsel, aggsel_env);
            Cor_TupleStack delta_stk, grping_stk, answer_stk;

            aggsel_env->clear_all();

            Cor_TupleIterator iter(this, *(aggsel->groupbylist), aggsel_env,
                NULL, endmark, NULL);

            for (;;) {
                Cor_StackMark stackmark3;
                Cor_Tuple *dbtuple = iter.get_next_tuple();
    
                // aggsel->groupbylist and aggsel->gb_agg_val correspond to
                // the fetched tuple.
                stackmark3.pop_to();
    
                if (iter.no_match() || (dbtuple == NULL))
                    break;

                delta_stk.push(UnknownHashValue,dbtuple);
            }

#ifdef    AGGSEL_VERBOSE
            if (Cor_exEnv->C_verbose_aggsel_default)
                cout << "Delta is:\n" << delta_stk << endl
                     << "--------\n";
    
    /*
            {
                aggsel_env->clear_all();
                Cor_TupleIterator iter(this, *(aggsel->groupbylist),
                aggsel_env, NULL, startmark, NULL);

                for (;;) {
                    Cor_StackMark stackmark3;
                    Cor_Tuple *dbtuple = iter.get_next_tuple();
        
                    stackmark3.pop_to();
        
                    if (iter.no_match() || (dbtuple == NULL))
                        break;
   
                    cout << "Got tuple " << *dbtuple << endl; 
                }
            }
    */
#endif

            for (j=0; j < delta_stk.count(); j++) {

                Cor_Tuple *delta_tuple = delta_stk.get_ith_tuple(j);

                if (!delta_tuple)
                    continue;

                Cor_Tuple *grp_tuple = NULL;
                grping_stk.empty();
                answer_stk.empty();

                aggsel_env->clear_all();
                Cor_unify_arg_lists(*(aggsel->arglist),
                        *(delta_tuple->_args), aggsel_env, NULL);

                  // If startmark is NULL, everything is in the delta
                if (startmark) {
                    Cor_TupleIterator iter(this, *(aggsel->groupbylist),
                          aggsel_env, NULL, NULL, startmark,
                          &(aggsel->indexnum));

                    for (;;) {
                        grp_tuple = iter.get_next_tuple();

                        if (iter.no_match() || !grp_tuple)
                            break;

                        grping_stk.push(UnknownHashValue,grp_tuple);
                    }
                }

                {
                    Cor_TupleIterator iter(this, *(aggsel->groupbylist),
                         // aggsel_env, NULL, startmark, endmark,
                            aggsel_env, NULL, startmark, NULL,
                            &(aggsel->indexnum));

                    for (;;) {
                        Cor_Tuple *grp_tuple = iter.get_next_tuple();

                        if (iter.no_match() || !grp_tuple)
                            break;

                        grping_stk.push(UnknownHashValue,grp_tuple);
                        delta_stk.delete_tuple(grp_tuple);
                    }
                }

#ifdef    AGGSEL_VERBOSE
                if (Cor_exEnv->C_verbose_aggsel_default)
                    cout << "Grouping is:\n" << grping_stk << endl
                         << "--------\n";
#endif

                Cor_ArgList *al = new Cor_ArgList(AGG_SELECTOR_ITER_ARITY);
                Cor_CArg *var = new Cor_CArg();
                var->set_var(0);

                Cor_CArg *a0, *a1, *tmparg;
                a0 = new Cor_CArg(NIL_ARG);
                a1 = new Cor_CArg(NIL_ARG);

                for (k=0; k < grping_stk.count(); k++) {

                    Cor_Tuple *tuple = grping_stk.get_ith_tuple(k);

                    aggsel_env->clear_all();
                    Cor_unify_arg_lists(*(aggsel->arglist),
                            *(delta_tuple->_args), aggsel_env, NULL);

                    if (Cor_subsumes_arg_lists(*(aggsel->groupbylist),
                            tuple->args(), aggsel_env, NULL)) {

                        tmparg = tuple->args().make_list();
                        tmparg->simplify_me(aggsel_env);
                        a0 = new Cor_CArg(tmparg,a0);

                        tmparg = aggsel->gb_agg_val->make_list();
                        tmparg->simplify_me(aggsel_env);
                        a1 = new Cor_CArg(tmparg,a1);
                    }
                }

                (*al)[0] = *a0; 
                (*al)[1] = *a1; 

                delete a0;
                delete a1;

                (*al)[2] = *var;

                Cor_BindEnv *env = new Cor_BindEnv(1);
                Cor_TupleIterator *it = 
                    new Cor_TupleIterator(aggsel->rel, *al, env);

                it->get_next_tuple();

                while (!it->no_match()) {

                    var->simplify_me(env);
                    var->hash();

                    if (!var->is_list) {
                        *(Cor_exEnv->err_strm)
                            << "Warning: Aggregate Selector "
                            << aggsel->rel->name
                            << " did not return a list: "
                            << *var << endl;
                        continue;
                    }        
          
                    var->unmake_lists(&answer_stk);

                    var->free_mem();
                    var->set_var(0);
                    it->get_next_tuple();
                }

                delete it;
                delete env;
                delete al;
                delete var;

#ifdef    AGGSEL_VERBOSE
                if (Cor_exEnv->C_verbose_aggsel_default)
                    cout << "Answers are:\n" << answer_stk << endl
                         << "--------\n";
#endif

                for (k=0; k < grping_stk.count(); k++) {

                    Cor_Tuple *tuple = grping_stk.get_ith_tuple(k);

                    if (! answer_stk.delete_eq_tuple(tuple)) {
                        // This tuple is not in the answer,
                        // remove it from the relation.
                        delete_id_tuple(this,tuple);
                        num_added--;
                    }
                }

                    // Answers to be inserted
                if (!answer_stk.is_empty()) {

                    int new_added = 0;    // New answers

                    for (k=0; k < answer_stk.count(); k++) {

                        Cor_Tuple *tuple = answer_stk.get_ith_tuple(k);

                          // Insert this answer into the relations
                        if (tuple)
                            if (!insert_new(tuple->args(),NULL))
                                delete tuple;
                            else
                                new_added++;
                    }

                    if (new_added) 
                        num_added = new_added;
                }
            }

            stack_mark2.pop_to();
            RELEASE(aggsel,aggsel_env)
        }
    }
    
    stack_mark1.pop_to();

#ifdef    AGGSEL_VERBOSE
    if (Cor_exEnv->C_verbose_aggsel_default)
        cout << "Got a total of " << num_added
             << " aggregated answers.\n";
#endif
    return(num_added);
}

// *************************************************************************
//  The following has to be fixed some time or the other, but is complicated
//  by structure sharing.
// *************************************************************************
void Cor_StorageRelation::free_all()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    free_allMarks();

    delete delta_relations;
    delta_relations = NULL;

    Cor_RelationLink *curr, *prev = NULL;
    for (curr = free_reln_list; curr; curr=curr->next) {
        delete prev;
        prev = curr;
    }
    delete prev;
    free_reln_list = NULL;
}

// *************************************************************************
void Cor_StorageRelation::free_indices()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}

// *************************************************************************
Cor_StorageRelation::~Cor_StorageRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    free_all();
}

// *************************************************************************
int Cor_StorageRelation::insert_in_context(Cor_GoalNode *g_node,
        Cor_GoalNode *parent_goal)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!check_subsum) {

        // first insert tuple in the for_ordsearch relation to index context
        for_ordsearch->insert_tuple(g_node);

        ASSERT((Cor_T_Stack->count != 0));
        ASSERT((Cor_T_Stack->top()->context != NULL));
        Cor_T_Stack->top()->context->add_new_goal(g_node, parent_goal);

        return(1);
    } else {

         // Wait and do the insert into for_ordsearch at the end
         // (since it cannot be cleanly deleted otherwise)
        ASSERT((Cor_T_Stack->count != 0));
        ASSERT((Cor_T_Stack->top()->context != NULL));
        Cor_T_Stack->top()->context->add_new_goal(g_node, parent_goal);
    }


    Cor_ContextNode *g_c_node = g_node->find_context_node();
    ASSERT((g_c_node != NULL));
    int g_node_number = g_c_node->node_number;
    Cor_GoalNode *lowest_marked = NULL;
    int lowest_marked_number = g_node_number + 1;

    Cor_BindEnv *g_node_env;

      // this is a hack because of some restrictions in
      // subsumes_arg_lists
    g_node_env = new Cor_BindEnv(0);

    // now the code to find matching tuples from for_ordsearch relation
    // for subsumption purposes
    Cor_StackMark stack_mark10;
    Cor_TupleIterator iter10(for_ordsearch, g_node->args(), g_node_env);

    for (;;) {
        Cor_StackMark stackmark20;
        Cor_Tuple *tuple20 = iter10.get_next_tuple();
        stackmark20.pop_to();
        // Undo any bindings done by get_next_tuple();
        if (iter10.no_match())
            break;

        ASSERT((tuple20 != NULL));

        // this can happen because g_node has already been
        // inserted in for_ordsearch relation
        //
        // No, it isn't being done anymore. -SPF
        //if (tuple20 == g_node)
        //  continue;

        Cor_ContextNode *tuple20_c_node =
            ((Cor_GoalNode *) tuple20)->find_context_node();
        ASSERT((tuple20_c_node != NULL));
        int tuple20_node_number = tuple20_c_node->node_number;

        if (Cor_subsumes_arg_lists(tuple20->args(), g_node->args(),
            iter10.tuple_env, NULL))
        {
            // we know that tuple20 subsumes g_node
            stackmark20.pop_to();
            // undo bindings done by subsumes_arg_lists

            if (g_node_number < tuple20_node_number) {
                // can delete the new inserted goal node
                Cor_T_Stack->top()->context->delete_goal(g_node);
                stack_mark10.pop_to();
                delete g_node_env;
                return(0);     // not inserted
            }
        }
        stackmark20.pop_to();

        if (Cor_subsumes_arg_lists(g_node->args(), tuple20->args(),
            g_node_env, iter10.tuple_env))
        {
            stackmark20.pop_to();
                // Undo bindings done by subsumes_arg_lists

            if ((tuple20_node_number <= g_node_number) &&
                    (!(((Cor_GoalNode *) tuple20)->is_marked())) )
            {
                // tuple20 is below g_node on context and can be deleted
                Cor_T_Stack->top()->context->delete_goal(
                            (Cor_GoalNode *) tuple20);

                // NOTE: might need to delete goal from the
                // for_ordsearch relation
                continue;
            }
            else if ((tuple20_node_number <= g_node_number) &&
                (((Cor_GoalNode *) tuple20)->is_marked()))
            {
                // have to find the lowest marked node subsumed by g_node
                if (tuple20_node_number < lowest_marked_number) {
                    lowest_marked_number = tuple20_node_number;
                    lowest_marked = (Cor_GoalNode *) tuple20;
                }
                continue;
            }
        }
        stackmark20.pop_to();
    }
    stack_mark10.pop_to();

    // have removed all lower unmarked goals subsumed by g_node and
    // found the lowest marked goal subsumed by g_node

    if (lowest_marked_number <= g_node_number) {
        // g_node subsumes something marked below it
        // hence, nodes need to be collapsed

        ASSERT((lowest_marked != NULL));
        Cor_T_Stack->top()->context->collapse_nodes(lowest_marked, parent_goal,
            g_node);
    }

    delete g_node_env;

    for_ordsearch->insert_tuple(g_node);

    return(1); // has been added to context but not subsumed
}

// *************************************************************************
