// *************************************************************************
//    Literal.h
// *************************************************************************

#ifndef _Literal_H_
#define _Literal_H_

#include "Predicate.h"
#include "Entity.h"
#include "ArgList.h"
#include "BitVector.h"

// *************************************************************************

enum pred_kind {
    COR_REGULAR,
    COR_MAGIC,
    COR_SUPMAGIC,
    COR_FACTORMAGIC,
    COR_DONE_MAGIC
};

// *************************************************************************

// *************************************************************************
// A Literal is an adorned predicate with some instantiation of its terms.
// NOTE : Literal is of variable size, since ArgList is of variable size.

class Cor_Literal {

  private:

  public:
    Cor_Predicate   *predicate;
    pred_kind        kind;

    // The pred,rule fields should be subsumed by rel !!
    Cor_CArg       pred;       // index into global preds[] for REG, MAG 
    Cor_Clause    *rule;       // rule, pos for COR_SUPMAGIC preds
    int            pos;
    Cor_Clause    *supp_rule;  // used in nested_scc rewriting ONLY

    char           base_pred;  // 1 if this is a base predicate
    char           negated;    // 1 if predicate is negated
    char           arith_pred; // 1 if predicate is arithmetic (=,>,<=, etc)

    Cor_BitVector  adorn;
    Cor_Relation  *rel;
    Cor_ArgList    args;

#ifdef    REUSE_MEMORY
    void* operator new(size_t size) {
        return(Coral_Free->acquire(size));
    }

    void  operator delete(void *trash, size_t size) {
        Coral_Free->dispose((Cor_Ent_Element*) trash, size);
    }
#endif

    Cor_Literal(Cor_ParserStruct *Cor_parserStruct, Cor_CArg& name,
            int arg_count);

    Cor_Literal(int arg_count);
   ~Cor_Literal();

    inline int arity()              { return args.count(); }
    inline Cor_CArg& name() const   { return predicate->name; }
};

// *************************************************************************

#endif _Literal_H_
