// *************************************************************************
//    RuleInfo.C
// *************************************************************************

#include <stdlib.h>
#include <string.h>

#include "RuleInfo.h"

#include "ArgList.h"
#include "StackMark.h"
#include "Config.h"
#include "Rule.h"
#include "Table.h"
#include "RMarkArray.h"
#include "TupleIterator.h"
#include "RelationList.h"
#include "SccInfo.h"
#include "BindEnv.h"
#include "Globals.h"
#include "Literal.h"
#include "Execute.h"
#include "CoralBreak.h"
#include "Debug.h"


// *************************************************************************

extern void Cor_AddProfileInfo(Cor_ProfileInfo *A, Cor_ProfileInfo *B);
extern int Cor_SimpleCopyArgs(Cor_ArgList *new_arglist,
        Cor_ArgList *old_arglist,
        Cor_BindEnv *rule_env, Cor_BindEnv *dont_rename_env,
        Cor_TermLink *&renamed_vars);
extern int Cor_backpatch_bindenvs(Cor_StackMark start,
        Cor_BindEnv *dont_rename_env,
        Cor_BindEnv *rule_env, Cor_TermLink *&renamed_vars);

extern void Cor_sprint_rule(Cor_Clause *rule, char *str);
void Cor_do_dump(Cor_Table *F, Cor_RuleInfo *rInfo, Cor_BindEnv *new_env,
    Cor_ArgList *new_arglist);

// *************************************************************************

#ifdef    DO_PROFILE

static char rule_buf[MAX_RULE_LENGTH];

#endif

// *************************************************************************

#ifdef DO_EXPLAIN
#ifdef DO_PROFILE
#define HANDLE_SUCCESS { \
    if ((rel->local_options & REL_EXPLAIN) && (Cor_T_Stack->count > 0))\
       Cor_do_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
        (pinfo->i_newfacts)++; \
    }
#else
#define HANDLE_SUCCESS  { \
    if ((rel->local_options & REL_EXPLAIN) && (Cor_T_Stack->count > 0)) \
       Cor_do_dump(F, rInfo, env, &(rInfo->arg_list[index])); \
              }
#endif
#else
#ifdef DO_PROFILE
#define HANDLE_SUCCESS  { (pinfo->i_newfacts)++; }
#else
#define HANDLE_SUCCESS  { }
#endif
#endif

// *************************************************************************
Cor_RuleInfo::Cor_RuleInfo(Cor_Clause *rul)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (rul->var_names)
        num_vars = rul->var_names->count();
    else 
        num_vars = 0;

    num_literals = rul->num_preds;
    arg_list = new Cor_ArgList[num_literals + rul->head_deletes_count + 1];
    relations = new int[num_literals + rul->head_deletes_count + 2];
    offsets = new int[num_literals + rul->head_deletes_count + 2];

    this->rule = rul;

#ifdef    DO_PROFILE
    Cor_sprint_rule(rul, rule_buf);
    char *buf = new char[strlen(rule_buf)+1];
    strcpy(buf, rule_buf);
    Cor_ProfileInfo *temp = new Cor_ProfileInfo(buf);
    pInfo = *temp;
    delete [] buf;
    delete temp;
#endif
}

// *************************************************************************
Cor_RuleInfo::~Cor_RuleInfo()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] arg_list;
    delete [] relations;
    delete [] offsets;
}

// *************************************************************************
//    Dump out information to be used by the Explain tool

void Cor_do_dump(Cor_Table *F, Cor_RuleInfo *rInfo, Cor_BindEnv *new_env,
    Cor_ArgList *new_arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ostream *strm = Cor_exEnv->explain_strm;

    if (!strm)
        return;

    for (int j=0; j <= rInfo->num_literals; j++) {
        if (j==rInfo->num_literals) {
            // First Argument
            *strm << "def_"
                  << F->moduledata->export_pred->name.symbol() << "(";
            rInfo->rule->head->predicate->name.print(new_env, strm);
            *strm << "(";
            ((Cor_ArgList *)new_arglist)->print_dump(new_env, strm);
            *strm << ")";
        } else {
            *strm << "use_"
                  << F->moduledata->export_pred->name.symbol() << "(";
            rInfo->rule->preds[j]->predicate->name.print(new_env, strm);
            *strm << "(";
            rInfo->arg_list[j].print_dump(new_env, strm);
            *strm << ")";
        }

        *strm << ", [";

        rInfo->rule->head->predicate->name.print(new_env, strm);
        *strm << "(";
        ((Cor_ArgList *)new_arglist)->print_dump(new_env, strm);
        *strm << ")";

        for (int i=0; i < rInfo->num_literals; i++) {
            *strm << ",";
            rInfo->rule->preds[i]->predicate->name.print(new_env, strm);
            *strm << "(";
            rInfo->arg_list[i].print_dump(new_env, strm);
            *strm << ")";
        }

        *strm << "]).\n";
    }

    // Dump info for two new relations
    for (int j1=0; j1 <= rInfo->num_literals; j1++) {
        if (j1==rInfo->num_literals) {
            // First Argument
            *strm << "def_new_"
                  << F->moduledata->export_pred->name.symbol() << "(";
            rInfo->rule->head->predicate->name.print(new_env, strm);
            *strm << "(";
            ((Cor_ArgList *)new_arglist)->print_dump(new_env, strm);
            *strm << ")";
        } else {
            *strm << "use_new_"
                  << F->moduledata->export_pred->name.symbol() << "(";
            rInfo->rule->preds[j1]->predicate->name.print(new_env, strm);
            *strm << "(";
            rInfo->arg_list[j1].print_dump(new_env, strm);
            *strm << ")";
        }

        *strm << ", [";

        rInfo->rule->head->predicate->name.print(new_env, strm);
        *strm << "(";
        ((Cor_ArgList *)new_arglist)->print_dump(new_env, strm);
        *strm << ")";

        for (int i1=0; i1 < rInfo->num_literals; i1++) {
            *strm << ",";
            rInfo->rule->preds[i1]->predicate->name.print(new_env, strm);
            *strm << "(";
            rInfo->arg_list[i1].print_dump(new_env, strm);
            *strm << ")";
        }

        *strm << "]).\n";
    }

    *strm << flush;
}

// *************************************************************************
