// *************************************************************************

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <limits.h>
#include <ctype.h>
#include "DebugSolvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#if defined(SOLAR_386) || defined(SOLAR) 
#ifndef RUSAGE_SELF
#include <fcntl.h>
#include <sys/procfs.h>
#endif
#endif

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "DerivedRelation.h"
#include "BuiltinUtil.h"
#include "GlobalBuiltinVars.h"
#include "Util.h"
#include "CoralBreak.h"
#include "QFMaterializedModuleData.h"

#include "Debug.h"
#include "New.h"

// *************************************************************************

static int   BP_Init = 0;         // Indicates a set_break has been done

static Cor_CcoralTimer coralTimer;

static int _CORAL_counter;

// *************************************************************************
//      Handle print_module
// *************************************************************************
Cor_BindEnv *Cor_PrintModuleSolver(Cor_BuiltinRelation &,
                                   Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int argcount = iterator.arg_list.count();

    if (argcount != 2) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "PrintModuleSolver");
        return (iterator.bindenv);
    }
    Cor_ArgList *copy_args = new Cor_ArgList(argcount);
    Cor_CopyArgs(copy_args, &iterator.arg_list,
                 iterator.bindenv, argcount, NULL);

    char arg_char[100];          // relation name
    char ad[100];                // adornment

    int pos = 0;
    ((*copy_args)[0]).sprint(arg_char, &pos);
    pos = 0;
    ((*copy_args)[1]).sprint(ad, &pos);

    // convert the adornment
    Cor_BitVector adornment;
    if (Cor_ConvertAdornment(adornment, ad, strlen(ad)) == -1) {
        *(Cor_exEnv->err_strm) << "Bad kind of adornment.\n";
        return (iterator.bindenv);
    }
    Cor_CArg rel_name = Cor_CArg(arg_char);
    if (!Cor_CurDB->RelationTable) {
        CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
        return (iterator.bindenv);
    }

    Cor_DerivedRelation *rel = (Cor_DerivedRelation *)
        Cor_CurDB->RelationTable->lookup(rel_name);

    if (!rel) {
        CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
        return (iterator.bindenv);
    }
    Cor_DerivedMethod *method;
    for (method = rel->methods; method; method = method->next) {
        if (!method->valid_method)
            continue;

        if (adornment == method->adornment) {
            Cor_QFMaterializedModuleData *md =
                    (Cor_QFMaterializedModuleData *) method->md;
            if (!md)
                CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
            else
                md->print(Cor_exEnv->out_strm);
            return (iterator.bindenv);
        }
    }

    CORAL_error(COR_MODULE_NOT_FOUND, NULL, "PrintModuleSolver");
    return (iterator.bindenv);
}

// *************************************************************************
//    Handle coral break
// *************************************************************************
Cor_BindEnv *Cor_BreakSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int argcount = iterator.arg_list.count();

    if (!BP_Init) {
        BP_Init = 1;
        Cor_brk_Info->all = 0;
        Cor_brk_Info->module_flag = 1;
        Cor_brk_Info->scc_flags = new Cor_BitVector(50);
        for (int j = 0; j <= 49; j++) {
            (Cor_brk_Info->rule_flags[j]) = NULL;
        }
    }
    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "BreakSolver");
        return (iterator.bindenv);
    }
    Cor_ArgList *copy_args = new Cor_ArgList(argcount);
    Cor_CopyArgs(copy_args, &iterator.arg_list,
                 iterator.bindenv, argcount, NULL);

    char arg_char[100];
    int pos = 0;
    int num = 0;

    ((*copy_args)[0]).sprint(Cor_brk_Info->module_name, &pos);
    strcat(Cor_brk_Info->module_name, "0");

    for (int i = 1; i <= argcount - 1; i++) {
        pos = 0;
        ((*copy_args)[i]).sprint(arg_char, &pos);
        int len = strlen(arg_char);
        if ((len > 3) && (strncmp(arg_char, "scc", 3) == 0)) {
            int index = 3;
            int done = 0;
            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index]))
                    done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return (0);
            } else {
                int s_num = atoi((char *) (arg_char + 3));
                Cor_brk_Info->scc_flags->set(s_num);
                num = s_num;
            }
        } else if ((len > 4) && (strncmp(arg_char, "rule", 4) == 0)) {
            int index = 4;
            int done = 0;

            if (!(Cor_brk_Info->rule_flags[num - 1]))
                (Cor_brk_Info->rule_flags[num - 1]) = new Cor_BitVector(50);

            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index]))
                    done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return (0);
            } else {
                int r_num = atoi((char *) (arg_char + 4));
                (Cor_brk_Info->rule_flags[num - 1])->set(r_num);
            }
        } else if ((len == 3) && (strcmp(arg_char, "all") == 0)) {
            Cor_brk_Info->all = 1;
        } else {
            //    BUILTIN_FAIL_RETURN
            return (0);
        }
    }
    return (iterator.bindenv);
}

// *************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

Cor_BindEnv *Cor_ClearBreakSolver(Cor_BuiltinRelation &,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int argcount = iterator.arg_list.count();

    if (!BP_Init) {
        *(Cor_exEnv->err_strm) << "No breakpoints are currently set.\n";
        //    BUILTIN_FAIL_RETURN
        return (0);
    }
    if (argcount <= 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "ClearBreakSolver");
        return (iterator.bindenv);
    }
    Cor_ArgList *copy_args = new Cor_ArgList(argcount);
    Cor_CopyArgs(copy_args, &iterator.arg_list,
                 iterator.bindenv, argcount, NULL);

    char arg_char[100];
    int pos = 0;
    int num = 0;

    ((*copy_args)[0]).sprint(Cor_brk_Info->module_name, &pos);
    strcat(Cor_brk_Info->module_name, "0");

    /* Only the module name is given */
    if (argcount == 1)
        Cor_brk_Info->module_flag = 0;

    for (int i = 1; i <= argcount - 1; i++) {
        pos = 0;
        ((*copy_args)[i]).sprint(arg_char, &pos);
        int len = strlen(arg_char);
        if ((len > 3) && (strncmp(arg_char, "scc", 3) == 0)) {
            int index = 3;
            int done = 0;
            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index]))
                    done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return (0);
            } else {
                int s_num = atoi((char *) (arg_char + 3));
                Cor_brk_Info->scc_flags->set(s_num, 0);
                num = s_num;
            }
        } else if ((len > 4) && (strncmp(arg_char, "rule", 4) == 0)) {
            int index = 4;
            int done = 0;

            if (!(Cor_brk_Info->rule_flags[num - 1]))
                continue;

            while ((arg_char[index] != '\0') && (!done)) {
                if (!isdigit(arg_char[index]))
                    done = 1;
                index++;
            }

            if (done) {
                //    BUILTIN_FAIL_RETURN
                return (0);
            } else {
                int r_num = atoi((char *) (arg_char + 4));
                (Cor_brk_Info->rule_flags[num - 1])->set(r_num, 0);
            }
        } else if ((len == 3) && (strcmp(arg_char, "all") == 0)) {
            Cor_brk_Info->module_flag = 0;
            if (Cor_brk_Info->all)
                Cor_brk_Info->all = 0;
            else {
                // Turn off all (like really all) breakpoints
                (Cor_brk_Info->scc_flags)->clear();
                int j;
                for (j = 0; j < 50; j++)
                    if (Cor_brk_Info->rule_flags[j])
                        (Cor_brk_Info->rule_flags[j])->clear();
            }
        } else {
            // BUILTIN_FAIL_RETURN
            return (0);
        }
    }
    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_ResetCounterSolver(Cor_BuiltinRelation &,
                                    Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.arity() != 1) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad arity for reset_c_counter()";
        iterator.set_no_match();
        return(NULL);
    }
    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_num()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad argument 1 to reset_c_counter =";
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        iterator.set_no_match();
        return(iterator.bindenv);
    }
    _CORAL_counter = make_int(term1.expr);

    return(iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_CounterSolver(Cor_BuiltinRelation &,
            Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.arity() != 1) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad arity for counter()";
        iterator.set_no_match();
        return (NULL);
    }

    Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term(term1);
    if (!term1.expr.is_var()) {
        *(Cor_exEnv->err_strm)
            << "CORAL :: Error -- bad argument 1 to counter =";
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        iterator.set_no_match();
        return (iterator.bindenv);
    }
    _CORAL_counter++;
    Cor_Term term2(_CORAL_counter, iterator.bindenv);
    Cor_unify_args(term2, term1);

    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_DebugSolver(Cor_BuiltinRelation & rel,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 0) {
        CORAL_error(COR_EMPTY_ARGLIST, NULL, "DebugSolver");
        return (iterator.bindenv);
    }
    if (rel.name.equals(Cor_SetSymbol)) {
        //Set each global variable specified in the argument list

        FOR_EACH_ARG(arg, iterator.arg_list) {
            Cor_Term term1(*arg, iterator.bindenv);
            Cor_deep_dereference_term(term1);
            if (!term1.expr.is_symbol()) {
                CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
                term1.printon(Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << endl;
                iterator.set_no_match();
                return (iterator.bindenv);
            }
            Cor_AssignValueToGlobal(term1.expr, 1);
        }
        END_EACH_ARG;
    } else if (rel.name.equals(Cor_ClearSymbol)) {
        //Clear each global variable specified in the argument list

        FOR_EACH_ARG(arg, iterator.arg_list) {
            Cor_Term term1(*arg, iterator.bindenv);
            Cor_deep_dereference_term(term1);
            if (!term1.expr.is_symbol()) {
                CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
                term1.printon(Cor_exEnv->err_strm);
                *(Cor_exEnv->err_strm) << endl;
                iterator.set_no_match();
                return (iterator.bindenv);
            }
            Cor_AssignValueToGlobal(term1.expr, 0);
        }
        END_EACH_ARG;
    } else if (rel.name.equals(Cor_AssignSymbol)) {
        //Assign the value in the second argument to the global variable 
        //specifed in the first argument

        if (iterator.arg_list.count() != 2) {
            CORAL_error(COR_BAD_ARITY, "2 arguments needed", "DebugSolver");
            return (iterator.bindenv);
        }

        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        if (!term1.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            iterator.set_no_match();
            return (iterator.bindenv);
        }

        Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term(term2);

        if (term2.expr.is_num()) {
            switch (term2.expr.Kind()) {

            case INT_ARG:
                Cor_AssignValueToGlobal(iterator.arg_list[0],
                                        iterator.arg_list[1].NumIval());
                break;

            case DOUBLE_ARG:
                Cor_AssignValueToGlobal(iterator.arg_list[0],
                                        (iterator.arg_list[1].NumDval()));
                break;

            default:
                break;
            }
        } else if (term2.expr.is_symbol()) {
            char *file_name = term1.expr.symbol();
            if (!strcmp(file_name, "trace_file")) {
                char *new_file = term2.expr.symbol();
                if (!set_trace_strm(new_file)) {
                    iterator.set_no_match();
                    return (iterator.bindenv);
                }
            } else if (!strcmp(file_name, "print_file")) {
                char *new_file = term2.expr.symbol();
                if (!set_out_strm(new_file)) {
                    iterator.set_no_match();
                    return (iterator.bindenv);
                }
            } else {
                CORAL_error(COR_BAD_GLOBAL_VAR, NULL, "AssignValueToGlobal");
                iterator.set_no_match();
                return (iterator.bindenv);
            }
        } else {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "DebugSolver");
            term2.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            iterator.set_no_match();
            return (iterator.bindenv);
        }
    } else {
        // Should never reach here. Safety check.
        abort();
    }

    return (iterator.bindenv);
}

// *************************************************************************
Cor_BindEnv *Cor_StartTimeSolver(Cor_BuiltinRelation &,
                                 Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    extern int Cor_beginMem;

#if  defined(SUN) || defined(LINUX)
    extern struct rusage Cor_SystemUsage;
    getrusage(RUSAGE_SELF, &Cor_SystemUsage);
#endif
#if defined(SOLAR_386) || defined(SOLAR) 
    #ifdef PIOCUSAGE
    int fdproc;
    char proc[15];
    extern prusage_t Cor_SystemPrUsage;
    sprintf(proc, "/proc/%d", getpid());
    if ((fdproc = open(proc,O_RDONLY)) == -1) {
       CORAL_error(COR_CANNOT_OPEN_FILE, proc, "StartTimeSolver");
    }
    if (ioctl(fdproc, PIOCUSAGE, &Cor_SystemPrUsage) == -1) {
       CORAL_error(COR_IOCTL_ERROR, "", "StartTimeSolver");
    }
    close(fdproc);
    #else 
    extern struct rusage Cor_SystemUsage;
    getrusage(RUSAGE_SELF, &Cor_SystemUsage);
    #endif 
#endif


    Cor_beginMem = (int) sbrk(0);        // save the top of the execution stack.

#ifdef HPUX
    coralTimer.reset();
#endif

    return (iterator.bindenv);
}

#define COR_U_SEC 1000000

// *************************************************************************
#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

Cor_BindEnv *Cor_EndTimeSolver(Cor_BuiltinRelation &,
                               Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CoralStream *fd;

    
    if (iterator.arg_list.count() == 1) {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        if (!term1.expr.is_symbol()) {
            *(Cor_exEnv->err_strm)
                << "CORAL :: Error -- bad argument to display_timer =";
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            iterator.set_no_match();
            return (iterator.bindenv);
        }
        char *file_name = term1.expr.symbol();

        if (!strcmp(file_name, "cout_strm"))
            fd = cout_strm;
        else if (!strcmp(file_name, "cerr_strm"))
            fd = cerr_strm;
        else if (!strcmp(file_name, Cor_exEnv->out_filename))
            fd = Cor_exEnv->out_strm;
        else if (!strcmp(file_name, Cor_exEnv->trace_filename))
            fd = Cor_exEnv->trace_strm;
        else if (!(fd = Cor_open_outfile(file_name, ios::app))) {
            CORAL_error(COR_CANNOT_OPEN_FILE, file_name, "EndTimeSolver");
            FAIL_RETURN;
        }
    } else
        fd = Cor_exEnv->err_strm;


#if    defined(SUN) || defined(LINUX) || defined(HPUX)
    extern struct rusage Cor_SystemUsage;
#    ifndef HPUX
    extern struct rusage Cor_startRUsage;
#    endif
#endif
#if    defined(SUN) || defined(LINUX) || defined(HPUX)
    struct rusage tmp_usage;
#endif
#if    defined(SUN) || defined(LINUX)

    long t1, t2;
    double t3, t4;

    getrusage(RUSAGE_SELF, &tmp_usage);
    t1 = tmp_usage.ru_utime.tv_sec - Cor_SystemUsage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec - Cor_SystemUsage.ru_utime.tv_usec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "User CPU Usage = " << t3 << " seconds\n";

    t1 = tmp_usage.ru_stime.tv_sec - Cor_SystemUsage.ru_stime.tv_sec;
    t2 = tmp_usage.ru_stime.tv_usec - Cor_SystemUsage.ru_stime.tv_usec;
    t4 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "System CPU Usage = " << t4 << " seconds\n";

    // t1 = tmp_usage.ru_utime.tv_sec - startRUsage.ru_utime.tv_sec;
    // t2 = tmp_usage.ru_utime.tv_usec - startRUsage.ru_utime.tv_usec;
    // t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
    // this was to show accumulative. changed to show total. -wei
    *fd << "Total CPU Usage = " << (t3+t4) << " seconds\n";
#elif defined(HP_UX)
    coralTimer.look();
    *fd << "User CPU Usage = " << coralTimer.usr << " seconds\n";
    *fd << "System CPU Usage = " << coralTimer.sys << " seconds\n";
    *fd << "Total CPU Usage = " << (coralTimer.usr + coralTimer.sys)
        << " seconds\n";
#endif

// the following stuff are quite buggy. -wei

#if    defined(SUN) || defined(LINUX) || defined(HPUX)
#ifdef DEBUG
    *fd << "Page Faults = "
        << ((tmp_usage.ru_minflt + tmp_usage.ru_majflt)
           - (Cor_SystemUsage.ru_minflt + Cor_SystemUsage.ru_majflt)) << endl;

    *fd << "Number of Context Switches = "
        << ((tmp_usage.ru_nvcsw + tmp_usage.ru_nivcsw)
           - (Cor_SystemUsage.ru_nvcsw + Cor_SystemUsage.ru_nivcsw)) << endl;
//    *fd << "Number of Swaps = "
//        << (tmp_usage.ru_nswap - Cor_SystemUsage.ru.nswap) << endl;
#endif
#endif

#if defined (SOLAR_386) || defined (SOLAR)
    #ifdef PIOCUSAGE
    int fdproc;
    char proc[15];
    prusage_t tmp_usage;
    extern prusage_t Cor_SystemPrUsage;
    sprintf(proc, "/proc/%d", getpid());
    if ((fdproc = open(proc,O_RDONLY)) == -1) {
       CORAL_error(COR_CANNOT_OPEN_FILE, proc, "EndTimeSolver");
    }
    if (ioctl(fdproc, PIOCUSAGE, &tmp_usage) == -1) {
       CORAL_error(COR_IOCTL_ERROR, "", "EndTimeSolver");
    }
    close(fdproc);
    long t1, t2;
    double t3, t4;
    t1 = tmp_usage.pr_utime.tv_sec - Cor_SystemPrUsage.pr_utime.tv_sec;
    t2 = tmp_usage.pr_utime.tv_nsec - Cor_SystemPrUsage.pr_utime.tv_nsec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "User CPU Usage = " << t3 << " seconds\n";

    t1 = tmp_usage.pr_stime.tv_sec - Cor_SystemPrUsage.pr_stime.tv_sec;
    t2 = tmp_usage.pr_stime.tv_nsec - Cor_SystemPrUsage.pr_stime.tv_nsec;
    t4 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "System CPU Usage = " << t4 << " seconds\n";

    *fd << "Total CPU Usage = " << (t3+t4) << " seconds\n";
    *fd << "Page Faults = "
        << ((tmp_usage.pr_minf + tmp_usage.pr_majf)
           - (Cor_SystemPrUsage.pr_minf + Cor_SystemPrUsage.pr_majf)) << endl;

    *fd << "Number of Context Switches = "
        << ((tmp_usage.pr_vctx + tmp_usage.pr_ictx)
           - (Cor_SystemPrUsage.pr_vctx + Cor_SystemPrUsage.pr_ictx)) << endl;
    *fd << "Number of Swaps = "
	<< (tmp_usage.pr_nswap - Cor_SystemPrUsage.pr_nswap) << endl;
    #else 
    extern struct rusage Cor_SystemUsage;
    struct rusage tmp_usage;
    getrusage(RUSAGE_SELF, &tmp_usage);

    long t1, t2;
    double t3, t4;
    t1 = tmp_usage.ru_utime.tv_sec - Cor_SystemUsage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec - Cor_SystemUsage.ru_utime.tv_usec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "User CPU Usage = " << t3 << " seconds\n";

    t1 = tmp_usage.ru_stime.tv_sec - Cor_SystemUsage.ru_stime.tv_sec;
    t2 = tmp_usage.ru_stime.tv_usec - Cor_SystemUsage.ru_stime.tv_usec;
    t4 = ((double) t1) + ((double) t2) / COR_U_SEC;
    *fd << "System CPU Usage = " << t4 << " seconds\n";

    *fd << "Total CPU Usage = " << (t3+t4) << " seconds\n";
    *fd << "Page Faults = "
        << ((tmp_usage.ru_minflt + tmp_usage.ru_majflt)
           - (Cor_SystemUsage.ru_minflt + Cor_SystemUsage.ru_majflt)) << endl;

    *fd << "Number of Context Switches = "
        << ((tmp_usage.ru_nvcsw + tmp_usage.ru_nivcsw)
           - (Cor_SystemUsage.ru_nvcsw + Cor_SystemUsage.ru_nivcsw)) << endl;
    *fd << "Number of Swaps = "
	<< (tmp_usage.ru_nswap - Cor_SystemUsage.ru_nswap) << endl;
    #endif
#endif


    if ((fd != Cor_exEnv->err_strm) && (fd != cout_strm)
        && (fd != Cor_exEnv->out_strm))
        delete fd;

    return (iterator.bindenv);
}

#undef FAIL_RETURN

// *************************************************************************
Cor_Tuple *Cor_CputimeSolver(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#if    defined(SUN) || defined(LINUX)
    extern struct rusage Cor_startRUsage;
#endif

    if (query == NULL) {
        *(Cor_exEnv->err_strm) << "Null argument passed to cputime \n";
        return (NULL);
    }
    int arity;
    if ((arity = query->arity()) != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to cputime needed", NULL);
        return (NULL);
    }
    if (!(query->args())[0].is_var()) {
        *(Cor_exEnv->err_strm)
           << "CORAL :: Error -- first argument to cputime should be free";
        (query->args())[0].printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        return (NULL);
    }
    double t3;

#if    defined(SUN) || defined(LINUX)
    long t1, t2;
    struct rusage tmp_usage;

// this time is the accumulative time spent in user mode.
    getrusage(RUSAGE_SELF, &tmp_usage);
    t1 = tmp_usage.ru_utime.tv_sec - Cor_startRUsage.ru_utime.tv_sec;
    t2 = tmp_usage.ru_utime.tv_usec - Cor_startRUsage.ru_utime.tv_usec;
    t3 = ((double) t1) + ((double) t2) / COR_U_SEC;
#else
    coralTimer.look();
    t3 = coralTimer.coral_usr;
#endif


    Cor_ArgList *alist = new Cor_ArgList(arity);
    (*alist)[0] = (t3);

    Cor_Tuple *return_tuple = new Cor_Tuple(alist);
    return (return_tuple);
}
#undef COR_U_SEC

// *************************************************************************
Cor_BindEnv *Cor_ShowMemSolver(Cor_BuiltinRelation&,
                Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    COUNT_MEM_USE
    cout << "\n-----Memory Use----\n";
    cout << "Memory Allocated: " << Coral_memory_alloc << endl;
    cout << "Memory Freed: " << Coral_memory_freed << endl;
    cout << "Memory in use: " << Coral_memory_inuse << endl;
    cout << "Max memory alloc so far :"<< Coral_max_memory_alloc << endl;
    cout << "Max memory inuse so far :"<< Coral_max_memory_inuse << endl;

    int j;
    for (j=0; j < 2000; j++) {
        if  (Coral_sizes_alloc[j]
                || Coral_sizes_inuse[j]
                || Coral_sizes_freed[j]) {
            cout << "Sizes: " << (j*8) << "->" << (j*8)+7 << endl;
                cout << "\tAlloc: " << Coral_sizes_alloc[j] << endl;
                cout << "\tFreed: " << Coral_sizes_freed[j] << endl;
                cout << "\tInUse: " << Coral_sizes_inuse[j] << endl;
        }
    }


    /*
    cout << "\n-----Closeup Memory Use (chunks)----\n";
    for (j=0; j < 8; j++) {
        cout << j << ":   Alloc: " << Coral_spec_sizes_alloc[j] << endl;
        cout << "\tFreed: " << Coral_spec_sizes_freed[j] << endl;
        cout << "\tInUse: " << Coral_spec_sizes_inuse[j] << endl;
    }
    */
    cout << "------------------\n\n";

#else
    cout << "show_mem: Memory allocate/free watching not compiled in." << endl;
#endif    COUNT_MEM_USE

    iterator.reset_no_match();
    return(NULL);
}

// *************************************************************************
