//**************************************************************************
//    FullMagic.C
//**************************************************************************
 
#include "FullMagic.h"

#include "SupMagic.h"

#include "RelationMisc.h"
#include "Literal.h"
#include "Globals.h"
#include "Unparse.h"
#include "SipOrder.h"

//**************************************************************************
// Assumptions:  Adorned rules (in the structure as_rules) have been sorted
// in sip order already.
//**************************************************************************

int opt_share_functors = 0;

static int debugmagic = 0;

// If set, suppress generation of magic literals for predicates with
// all arguments free.
static int suppress_allfree = 0;

//**************************************************************************
void nomagic(Cor_Collection *rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k, nrules;
    Cor_Clause *magicrule;
    extern Cor_Collection *magic_rules;    // Main.C
    Cor_Clause *rule;

    magic_rules->empty();
    nrules = rules->count();

    for (k=0; k < nrules; k++) {
        rule = rules->get_ith(k);
        magicrule = Cor_copy_rule(rule);
        magicrule->head = rule->head;
        Cor_alloc_preds(magicrule, rule->num_preds );
        for (j=0; j<rule->num_preds; ++j)
            magicrule->preds[j] = rule->preds[j];
        magic_rules->push(magicrule);
    }
}

//**************************************************************************
void fullmagic(Cor_Collection *ad_rules, Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *adrule;
    int j, nrules;

    nrules = ad_rules->count();

    for (j=0; j < nrules; j++) {
        adrule = ad_rules->get_ith(j);
          // Construct the magic rules
        if (0 && Cor_parserStruct->CurModule->NestedSccEval && adrule->supp)
            continue;

        do_magic_on_rule(adrule, magic_rules);
    }
}

//**************************************************************************
/*
    This is used to compute the number of "done_m_" literals to be added
    in the magic rule.  There are two cases:
        (1) If there is aggregation in the head, then pre-processing
            guarantees that only one "done_m_" literal needs to be 
            added to the rule.
        (2) If there is genuine negation in the body of the rule, add
            as many "done_m_" as there are negated literals.
*/

int compute_deriv_negated(Cor_Clause *adrule)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int ret_val = 0;

    //    For monotonic programs, there is no need to add any done_m_ literals.
    if (Cor_parserStruct->CurModule->Monotonic)
        return ret_val;

    if (adrule->head->args.has_grouping_args())
        ret_val = 1;
    else {
        //    There is genuine negation in the head of the rule
        for (int i=0; i < adrule->num_preds; ++i) {
            if (adrule->preds[i]->negated && (!adrule->preds[i]->base_pred)) {
                // negated and derived predicates
                ++ret_val;
            }
        }
    }
    return ret_val;
}

//**************************************************************************
void do_magic_on_rule(Cor_Clause *adrule, Cor_Collection *magic_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i,j;
    Cor_Clause *magicrule;

    if (debugmagic) {
        *cerr_strm << "Original rule :\n" << *adrule;
    }

    // NOTE: The following code is for Ordered Search.
    // It finds the number of derived negated literals in the body of 
    // a rule and hence the number of "done_m_" literals to be added
    // Also, it always uses ordered search whenever a derived negated
    // literal is encountered.

    int num_deriv_negated = compute_deriv_negated(adrule);
    // number of negated derived literals in adrule

    for (i=0; i < adrule->num_preds; ++i)
        create_derived_magic_rules(adrule, magic_rules, i);

    //    Construct the modified rule
    magicrule = Cor_copy_rule(adrule);

    if (! (allargsfree(& adrule->head->adorn) && suppress_allfree)) {
        int num_bound_functor_args = 0;
        if (opt_share_functors) {
            // Optimization to let the head share functors with the magic
            // predicate in the body
            num_bound_functor_args = count_bound_functor_args(adrule->head);

            // Create new variables for functor args
            make_new_rule_vars(magicrule, num_bound_functor_args);
        }
        Cor_alloc_preds(magicrule, adrule->num_preds + 1 + num_deriv_negated +
            num_bound_functor_args);
        magicrule->head = build_head_pred(magicrule, adrule->head);
        magicrule->preds[0] = build_magic_pred(magicrule, adrule->head,
            opt_share_functors);
        if (opt_share_functors) {
            create_functor_literals(magicrule);
        }
        for (j=num_bound_functor_args+1,i=0; i<adrule->num_preds; ++j, ++i) {
            // NOTE: The following code is for Ordered Search
            if ((Cor_parserStruct->CurModule->UseOrdSearch) &&
                (!Cor_parserStruct->CurModule->Monotonic) &&
                ((adrule->head->args.has_grouping_args()) ||
                (adrule->preds[i]->negated)) &&
                (!adrule->preds[i]->base_pred))
            {
                magicrule->preds[j] = build_magic_pred(magicrule,
                    adrule->preds[i], 0) ;
                magicrule->preds[j++]->kind = COR_DONE_MAGIC;
            }
            magicrule->preds[j] = adrule->preds[i];
         }
    }
    else {
        //    Copy out the original rule
        magicrule->head = adrule->head;
        Cor_alloc_preds(magicrule, adrule->num_preds + num_deriv_negated);
        for (j=0, i=0; i<adrule->num_preds; ++j,++i) {
            // NOTE: The following code is for Ordered Search
            if ((Cor_parserStruct->CurModule->UseOrdSearch) &&
                (!Cor_parserStruct->CurModule->Monotonic) &&
                (adrule->head->args.has_grouping_args() ||
                (adrule->preds[i]->negated)) &&
                (!adrule->preds[i]->base_pred))
            {
                magicrule->preds[j] = build_magic_pred(magicrule,
                    adrule->preds[i],0) ;
                magicrule->preds[j++]->kind = COR_DONE_MAGIC;
            }
            magicrule->preds[j] = adrule->preds[i];
        }
    }

    magicrule->head_deletes_count = adrule->head_deletes_count;
    magicrule->head_deletes = adrule->head_deletes;
    magic_rules->push(magicrule);
}

//**************************************************************************
/*
    Generate a magic rule for each derived right-hand side predicate that
    has an incoming arc in the sip.   Currently handles only linear sips.
*/

void create_derived_magic_rules(Cor_Clause *adrule,
    Cor_Collection *magic_rules, int pred_index)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j,p;
    Cor_Clause *magicrule;
    Cor_Sip *s;

    if (adrule->preds[pred_index]->base_pred) return;

    if (adrule->sips->sip_list)
        s = findadsip(adrule, pred_index);
    else
        s = NULL;
    
    //    Copy proto header
    magicrule = Cor_copy_rule(adrule);
    magicrule->head = build_magic_pred(adrule, adrule->preds[pred_index], 0);
    if (s) {
        Cor_alloc_preds(magicrule, numpredsinsip(&s->tail));
        for(j=0,p=firstpredinsip(&s->tail); p >= 0;
            p = nextpredinsip(&s->tail,p) )
        {
            if (p==0) {
                //    Head predicate
                if (!(allargsfree(& adrule->head->adorn) && suppress_allfree)) {
                    magicrule->preds[j++] =
                        build_magic_pred(adrule, adrule->head, 0);
                }
            }
            else {
                magicrule->preds[j++] =    adrule->preds[p-1];
            }
        }
        magicrule->num_preds = j;
    }
    else if (!(allargsfree( &adrule->preds[pred_index]->adorn) &&
        suppress_allfree))
    {
        magicrule->num_preds = 0;
    }
    else
        return;
    
    if (!(allargsfree( & magicrule->head->adorn) && 
            suppress_allfree))
    {
        magicrule->type = COR_MAGICRULE;
        magicrule->source_rule = adrule;
        magicrule->source_pred = adrule->preds[pred_index];

        if (magicrule->source_pred->negated ||
            adrule->head->args.has_grouping_args())
        {
            if (magicrule->clause_names->is_empty()) {
                
                  // Need to create a new name
                magicrule->clause_names->add(*Cor_DontCareSymbol,1);
            }
            magicrule->clause_names->set_negated(0,1);
        }

        magic_rules->push(magicrule);
        
        if (debugmagic) {
            *cerr_strm << " Magic rule:\n" << *magicrule << endl;
        }
    }
    else if (debugmagic)
        *cerr_strm << "No magic rule generated here-null rule body\n";
}

//**************************************************************************
