// *************************************************************************
//    Print.C
// *************************************************************************

#include "Print.h"

#include "Globals.h"
#include "Rule.h"
#include "Literal.h"
#include "Collection.h"
#include "Unparse.h"
#include "ExportInfo.h"

// *************************************************************************

static void print_status(Cor_Clause *rule, ostream *out);

// *************************************************************************
void print_bits(Cor_BitVector b, int flag, ostream *outf) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    register int i;
    char c[32];
    char ch;

    if (b.length == 0) return;
    *outf << "_";
    ch = (flag==0) ? 'd':'_';

    for (i=0; i<b.length; ++i)
        c[i] = b[i] ? ch : 'n';

    c[b.length]='\0';

    *outf << c;
}

// *************************************************************************
void print_bits(Cor_BitVector b, int flag)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    print_bits(b, flag, cerr_strm);
}

// *************************************************************************
static void print_status(Cor_Clause *rule, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *str;
    switch(rule->factor_status) {
        case 0:  str = "COR_NON_RECURSIVE   "; break;
        case 1:  str = "COR_RT_LINEAR       "; break;
        case 2:  str = "COR_LT_LINEAR       "; break;
        case 3:  str = "PSEUDO_LT_LINEAR    "; break;
        case 4:  str = "COR_MULTI_LINEAR    "; break;
        default: str = "COR_NON_LINEAR      "; break;
    }
    *out << str << " : ";
}

// *************************************************************************
void print_factor(Cor_Collection *factor_rules, ostream *outf,
     int show_status,Cor_Collection *org_rules )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (factor_rules->count() == 0) {
        *cerr_strm << "CORAL:: WARNING! Magic/Factoring for "
           << Cor_parserStruct->CurModule->name.symbol()
           << " results in module with no rules\n";
        return;
    }

    if (! Cor_parserStruct->CurModule->name.is_empty()) {
        *outf << "module "
              << Cor_parserStruct->CurModule->name.symbol() << ".\n";
    }
    for (Cor_ExportInfo *export_info=Cor_parserStruct->CurModule->ExportedPreds;
        export_info != NULL; export_info = export_info->next)
    {
        *outf << "export " << export_info->pred_name.symbol() << "[";
        if (! export_info->adorn_symbol.is_empty()) {
            *outf << export_info->adorn_symbol.symbol();
        }
        *outf << "].\n";
    }

    Cor_print_Annotations(*(Cor_parserStruct->CurModule), outf,
        Cor_parserStruct->make_index_annotations,
        Cor_parserStruct->agg_sel_annotations,
        Cor_parserStruct->AllowedAdornList,
        Cor_parserStruct->MultisetAnnotations,
        Cor_parserStruct->diskrel_annotations,
        Cor_parserStruct->asyncrel_annotations);

    Cor_Clause *factorrule;
    int j, nrules;

    nrules = factor_rules->count();

    for (j=0; j < nrules; j++) {
        factorrule = factor_rules->get_ith(j);

        if (show_status)
            print_status(factorrule, outf);

        *outf << *factorrule;
    }

    if (org_rules) { // added to print orginal rules
      *outf << "% original program:\n";

      Cor_Clause *orgrule;

      nrules = org_rules->count();

      for (j=0; j < nrules; j++) {
          orgrule = org_rules->get_ith(j);

          if (show_status)
              print_status(orgrule, outf);
          orgrule->sips = NULL;
          *outf << *orgrule;
      }
    }

    *outf << "end_module.\n\n";
}

// *************************************************************************
